/*
 * resume.c - Functions for waking devices up.
 *
 * Copyright (c) 2003 Patrick Mochel
 * Copyright (c) 2003 Open Source Development Labs
 *
 * This file is released under the GPLv2
 *
 */

#include <linux/device.h>
#include <linux/delay.h>
#include "power.h"


extern int sysdev_resume(void);


/**
 *	resume_device - Restore state for one device.
 *	@dev:	Device.
 *
 */

int resume_device(struct device * dev)
{
	if (dev->bus && dev->bus->resume)
		return dev->bus->resume(dev);
	return 0;
}

#if defined(CONFIG_PM_DEBUG) && defined(CONFIG_ARCH_FIONA)

#include <asm/hardware.h>
#include <asm/arch/pxa-regs.h>
static void fdputs(char *s)
{
    int     size = 0;
    int     x = 0, wait = 0;
    char    *cur = s;

    //return;

    if (s == NULL)
        return;
    size = strlen(s);

    // Print the string, waiting for each character to shift out before
    // writing the next one...
    for (x=0; x<size; x++)  {
        STTHR = *cur++;
        wait = 0;
        while ((STLSR & 0x20) == 0) {
            msleep(1);
            if (wait++ > 10) {
                STTHR = '^';
                return;
            }
        }
    }
}
#endif

void dpm_resume(void)
{
#if defined(CONFIG_PM_DEBUG) && defined(CONFIG_ARCH_FIONA)
    char theStr[255];
    theStr[0] = '\0';
#endif

	down(&dpm_list_sem);
	while(!list_empty(&dpm_off)) {
		struct list_head * entry = dpm_off.next;
		struct device * dev = to_device(entry);

		get_device(dev);
		list_del_init(entry);
		list_add_tail(entry, &dpm_active);

		up(&dpm_list_sem);
		if (!dev->power.prev_state)
        {
#if defined(CONFIG_PM_DEBUG) && defined(CONFIG_ARCH_FIONA)
            if ((dev->driver) && (dev->driver->name)) {
                sprintf(theStr,"Resuming device: %s\n",dev->driver->name);
                fdputs(theStr);
            }
            else {
                sprintf(theStr,"Resuming device no-name...\n");
                fdputs(theStr);
            }
#endif
			resume_device(dev);
        }
#if defined(CONFIG_PM_DEBUG) && defined(CONFIG_ARCH_FIONA)
        else  {
            sprintf(theStr,"Did not resume device %s - power.prev_state (%d) not zero\n",dev->driver->name,dev->power.prev_state);
            fdputs(theStr);
        }
#endif
		down(&dpm_list_sem);
		put_device(dev);
	}
	up(&dpm_list_sem);
}


/**
 *	device_resume - Restore state of each device in system.
 *
 *	Walk the dpm_off list, remove each entry, resume the device,
 *	then add it to the dpm_active list.
 */

void device_resume(void)
{
	down(&dpm_sem);
	dpm_resume();
	up(&dpm_sem);
}

EXPORT_SYMBOL_GPL(device_resume);


/**
 *	device_power_up_irq - Power on some devices.
 *
 *	Walk the dpm_off_irq list and power each device up. This
 *	is used for devices that required they be powered down with
 *	interrupts disabled. As devices are powered on, they are moved to
 *	the dpm_suspended list.
 *
 *	Interrupts must be disabled when calling this.
 */

void dpm_power_up(void)
{
	while(!list_empty(&dpm_off_irq)) {
		struct list_head * entry = dpm_off_irq.next;
		struct device * dev = to_device(entry);

		get_device(dev);
		list_del_init(entry);
		list_add_tail(entry, &dpm_active);
		resume_device(dev);
		put_device(dev);
	}
}


/**
 *	device_pm_power_up - Turn on all devices that need special attention.
 *
 *	Power on system devices then devices that required we shut them down
 *	with interrupts disabled.
 *	Called with interrupts disabled.
 */

void device_power_up(void)
{
	sysdev_resume();
	dpm_power_up();
}

EXPORT_SYMBOL_GPL(device_power_up);


