/**********************************************************************
 * Philips ISP176x Hardware Access Interface header file
 *
 * (c) 2002 Koninklijke Philips Electronics N.V. All rights reserved. <usb.linux@philips.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * File Name:   hal_intf.h
 *
 * History:     
 *
 * Date                Author                  Comments
 * ---------------------------------------------------------------------
 * Nov 29 2005        Prabhakar Kalasani      Initial Creation     
 *               
 ***********************************************************************/

/* Based on hal_intf.h */

#ifndef __HAL_INTF_H__
#define __HAL_INTF_H__


//#define DMA
/* Registers accessed by Interface functions */
/***** HC Registers *****/

#define		HC_REVISION_REG			0x00	/* 8-31Revision Register (read only) */
#define		HC_CAPABILITY_REG		0x00	/* 0-7 Capability Register (read only) */

#define		HC_SPARAMS_REG			0x04	/* Structural Parameters Register */
#define		HC_CPARAMS_REG			0x08	/* Capability Parameters Register */

#define		HC_SP_PORTROUTE1_REG	0x0C	/* Companion Port Route Register 1 */
#define		HC_SP_PORTROUTE2_REG	0x10	/* Companion Port Route Register 1 */
#define		HC_USBCMD_REG			0x20	/* USB Command Register */
#define		HC_USBSTS_REG			0x24	/* USB Status Register */
#define		HC_INTERRUPT_REG_EHCI		0x28	/* INterrupt Enable Register */
#define		HC_FRINDEX_REG			0x2C	/* Frame Index Register */

#define		HC_CTRLDSSEGMENT_REG	0x30	/* 4G Segment Selector Register (Only for 64bit)*/
#define		HC_PERIODICBASE_REG		0x34	/* Periodic List Base Address Register */
#define		HC_ASYNCLISTADDR_REG	0x38	/* Next Asynchronous List register */
#define		HC_CONFIGFLAG_REG		0x60	/* Conigured Flag  Register */

#define		HC_PORTSC1_REG		0x64	/* Port Status Control for Port1 */
#define		HC_PORTSC2_REG		0x68	/* Port Status Control for Port2 */
#define		HC_PORTSC3_REG		0x6C	/* Port Status Control for Port3 */
#define		HC_PORTSC4_REG		0x70	/* Port Status Control for Port4 */

//ISO Transfer Registers
#define		HC_ISO_PTD_DONEMAP_REG			0x130	/* ISO PTD Done Map Register */
#define		HC_ISO_PTD_SKIPMAP_REG			0x134	/* ISO PTD Skip Map Register */
#define		HC_ISO_PTD_LASTPTD_REG			0x138	/* ISO PTD Last PTD Register */

//INT Transfer Registers
#define		HC_INT_PTD_DONEMAP_REG			0x140	/* INT PTD Done Map Register */
#define		HC_INT_PTD_SKIPMAP_REG			0x144	/* INT PTD Skip Map Register */
#define		HC_INT_PTD_LASTPTD_REG			0x148	/* INT PTD Last PTD Register  */

//ATL Transfer Registers
#define		HC_ATL_PTD_DONEMAP_REG			0x150	/* ATL PTD Last PTD Register  */
#define		HC_ATL_PTD_SKIPMAP_REG			0x154	/* ATL PTD Last PTD Register  */
#define		HC_ATL_PTD_LASTPTD_REG			0x158	/* ATL PTD Last PTD Register  */

#define		HC_ATL_PTDDONE_THRESHOLDCNT_REG			0x350	/* Revision Register (read only) */
#define		HC_ATL_PTDDONE_THRESHOLDTIMEOUT_REG		0x351	/* Control Register */



//General Purpose Registers
#define		HC_HW_MODE_REG			0x300	/* H/W Mode Register  */
#define		HC_CHIP_ID_REG			0x304	/* Chip ID Register */
#define		HC_SCRATCH_REG			0x308	/* Scratch Register */
#define		HC_RESET_REG			0x30C	/* HC Reset Register */
										   
//Interrupt Registers
#define		HC_INTERRUPT_REG			0x310	/* Interrupt Register */
#define		HC_INTENABLE_REG			0x314	/* Interrupt enable Register */
#define		HC_ISO_IRQ_MASK_OR_REG		0x318	/* ISO Mask OR Register */
#define		HC_INT_IRQ_MASK_OR_REG		0x31C	/* INT Mask OR Register*/
#define		HC_ATL_IRQ_MASK_OR_REG		0x320	/* ATL Mask OR Register */
#define		HC_ISO_IRQ_MASK_AND_REG		0x324	/* ISO Mask AND Register */
#define		HC_INT_IRQ_MASK_AND_REG		0x328	/* INT Mask AND Register */
#define		HC_ATL_IRQ_MASK_AND_REG		0x32C	/* ATL Mask AND Register */

//RAM Registers
#define		HC_DMACONFIG_REG			0x330	/* DMA Config Register */
#define		HC_TRANS_COUNT_REG			0x334	/* Transfer Counter Register */
#define		HC_BUFFER_STAT_REG			0x338	/* Buffer status Register */
#define		HC_MEM_READ_REG				0x33C	/* Memory Register */

/*interrupt count and buffer status register*/

#define      HC_BUFFER_STATUS_REG		0x334
#define      HC_INT_THRESHOLD_REG    0x340 

#if defined(CONFIG_ARCH_FIONA)
#define		HC_PORT1_CONTROL_REG		0x374	/* ISP176x Port 1 Control Register */
#else
//OTG Registers
#define		HC_OTG_TIMER1_REG				0x370	/* OTG Timer 1 Register (read only) */
#define		HC_OTG_TIMER2_REG				0x374	/* OTG Timer 2 Register */
#define		HC_OTG_STATUS1_REG				0x378	/* OTG Status 1 Register */
#define		HC_OTG_STATUS2_REG				0x37C	/* OTG Status 2 Register */
#define		HC_OTG_SC_REG					0x380	/* OTG Status Change Register */
#define		HC_OTG_SC_IE_REG				0x384	/* OTG Status Change Interrupt Enable Register */
#endif

/* HC Registers Bit positions definitions */

/* 0x020  HC_USBCMD_REG			USBCMD (USB Command) register	*/
#define RS			0x00000001
#define	HCRESET			0x00000002
#define	LHCR			0x00000080

/* 0x300  HC_HW_CNFG_REG (16 bit) */
#define HC_GlOBALINT_ENABLE	0x0001		/* Bit 0    */
#define HC_INT_EDGE_TRIGGER	0x0002		/* Bit 1    */
#define HC_INT_ACTIVE_HIGH	0x0004		/* Bit 2    */
#define HC_DREQ_POLARITY_HIGH	0x0020		/* Bit 5	*/
#define	HC_DACK_POLARITY_HIGH	0x0040		/* Bit 6	*/
#define	HC_EOT_POLARITY_HIGH	0x0080		/* Bit 7	*/
#define HC_OTG_INTERRUPT        0x0400

/* 0x308  HC_SCRATCH_REG		HcScratch register		*/
#define	USB_RUNNING_SIGNATURE	0x13579BDF
#define	USB_SUSPEND_SIGNATURE	0x11235813

/* 0x310  HC_INTERRUPT_REG		HcInterrupt register		*/
#define	SOFITLINT		0x00000002
#define HCSUSP			0x00000020
#define CLKREADY		0x00000040
#define INT_IRQ			0x00000080
#define ATL_IRQ			0x00000100
#define ISO_IRQ			0x00000200
#define OTG_IRQ			0x00000400

#define DMA_CHANNEL1_PCI_ADDRESS                                0x98
#define DMA_CHANNEL1_LOCAL_ADDRESS                              0x9c
#define DMA_CHANNEL1_TRANSFER_SIZE                              0xa0
#define DMA_CHANNEL1_DESCRIPTOR_POINTER                 0xa4
#define DMA_CHANNEL1_THRESHOLD                                  0xb0
#define DMA_CHANNEL1_COMMAND_STATUS                             0xa9

#endif
