/******************************************************************
 *
 *  File:   ioc.h
 *
 *  Author: Nick Vaccaro <nvaccaro@lab126.com>
 *
 *  Date:   12/07/05
 *
 * Copyright 2005, Lab126, Inc.  All rights reserved.
 *
 *  Description:
 *     Constants and definitions for IOC
 *
 ******************************************************************/

#ifndef _IOC_H_
#define _IOC_H_

#undef LOCAL_CONFIG_PM

#include <linux/autoconf.h>
#ifdef __KERNEL__
#include <linux/miscdevice.h>
#include <linux/timer.h>
#endif
#include <linux/pn_lcd.h>

/****************************************
 *
 * IOC Command related constants
 *
 ****************************************/

//////////////////////////////////////
// Various Debug-related switches
//////////////////////////////////////
#undef DEBUG_WAKE_BUG

#undef DEBUG_IOC_COMMAND_RECEIVE
#undef DEBUG_IOC_FIFOS
#undef DEBUG_RX_BUG             // Debugging around IOC bug where IOC stops responding
#undef DEBUG_CORE_IO            // Print JUST TX and RX data to/from UART


#undef DEBUG_IOC_COMMAND_SEND
#undef DEBUG_SERIAL_INTERFACE
#undef DEBUG_VERBOSE            // Extra verbosity for tx / rx 
#undef DEBUG_BATTERY            // Debug the battery stuff
#undef DEBUG_PNLCD              // Debug the PNLCD stuff

#undef ALLOW_LOW_BATTERY_TO_WAKE_US // Define to allow low battery alerts to wake us up...

//
// Turn on general PDEBUG if any of the debug modes is enabled
//
#if defined(DEBUG_SERIAL_INTERFACE) || defined(DEBUG_IOC_COMMAND_SEND) || \
    defined(DEBUG_IOC_COMMAND_RECEIVE) || defined(DEBUG_IOC_FIFOS)
#define PDEBUG(args...) printk(args)
#define DEBUG_PNLCD
#else
#define PDEBUG(args...) do {} while(0)
#endif

#if defined(DEBUG_PNLCD)
#define NDEBUG(args...) printk(args)
#else
#define NDEBUG(args...) do {} while(0)
#endif

#ifdef DEBUG_BATTERY
#define BDEBUG(args...) printk(args)
#else
#define BDEBUG(args...) do {} while(0)
#endif

#ifdef DEBUG_WAKE_BUG
#define DEBUG_IOC_COMMAND_RECEIVE
#define DEBUG_IOC_FIFOS
#define DEBUG_RX_BUG             // Debugging around IOC bug where IOC stops responding
#define DEBUG_CORE_IO            // Print JUST TX and RX data to/from UART
#define wakelog(x...) printk(x)
#else
#define wakelog(x...) do {} while(0)
#endif

#

// Handy macros
#define memclr(a,l) memset(a,0,l)
#define MIN(x,y)    ((x<y)?x:y)
#define MAX(x,y)    ((x>y)?x:y)

#define VC_TIMERCONT() add_timer(&vc_timer)
#define VC_TIMEROFF()  del_timer_sync(&vc_timer)
#define VC_TIMERON(t)  mod_timer(&vc_timer, jiffies + (t))


// Used to indicate a place holder for command byte to be inserted later
#define PLACE_HOLDER    0
#define ENABLE  1
#define DISABLE 0

#define IOC_MAJOR                           10             // ioc device major number
#define IOCTL_IOC_MAGIC_NUMBER              'i'
#define IOC_DRIVER_NAME                     "ioc"
#define IOC_DRIVER_PATH                     "/dev/misc/ioc"
#define IOC_PROC_PARENT_NAME                "ioc"
#define IOC_PROC_ENTRY_NAME                 "ioc"
#define IOC_PROC_BATTERY_ENTRY_NAME         "battery"
#define IOC_PROC_VC_ENTRY_NAME              "vc"
#define IOC_PROC_ALT_BATTERY_ENTRY_NAME     "altbattery"
#define IOC_PROC_CLEAR_AMV_ENTRY_NAME       "camv"
#define IOC_PROC_AC_ENTRY_NAME              "ac"
#define IOC_PROC_DEBUG_ENTRY_NAME           "debug"
#define IOC_PROC_VDUMP_DEBUG_ENTRY_NAME     "vdump"
#define IOC_PROC_ALLOW_SLEEP_ENTRY_NAME     "allow_sleep"
#define IOC_PROC_RESET_ENTRY_NAME           "reset"
#define IOC_PROC_RESET_VTBL_ENTRY_NAME      "reset_vtbl"
#define IOC_PROC_LOCK_ENTRY_NAME            "lock"
#define IOC_PROC_CALC_BR_ENTRY_NAME         "calc_ohms"
#define IOC_PROC_VERSION_ENTRY_NAME         "version"
#define IOC_PROC_FLASH_FW_ENTRY_NAME        "flash_fw"
#define IOC_PROC_UP_LAMBDA_ENTRY_NAME       "up_lambda"
#define IOC_PROC_DOWN_LAMBDA_ENTRY_NAME     "down_lambda"
#define IOC_PROC_MIN_IMPEDANCE_ENTRY_NAME   "min_impedance"
#define IOC_PROC_MAX_IMPEDANCE_ENTRY_NAME   "max_impedance"
#define IOC_PROC_ICON_IMPEDANCE_ENTRY_NAME  "icon_impedance"
#define IOC_PROC_ICON_ALG_ENTRY_NAME        "icon_alg"
#define IOC_PROC_WATERMARK_ALG_ENTRY_NAME   "watermark_alg"
#define IOC_PROC_FAIL_TRANSITION_ENTRY_NAME "fail_next_transition"

#define BATTERY_INFO_STRING_SIZE            70          // Max size formatted Battery String
#define IOC_FW_FLASH_SIZE                   4096        // Size of IOC Flash
#define IOC_MAX_SIZE_FW_TEXT                (4096*8)    // Max Size of IOC FW text file
#define BATTERY_NOT_PRESENT_RAW_TEMP        0xff        // Temp == 0xff indicates NO BATTERY
#define IOC_CRITICAL_BATTERY_CAPACITY       10          // 10% for now...
#define IOC_GLOBALS_SIGNATURE               (0x696F6320)// 'ioc '
#define MAX_RX_PACKETS_TO_BUFFER            10          // Max packets to buffer off
                                                        // Buffer is between IRQ handler and BH
#define MAX_IOC_RX_PACKET_SIZE              8           // Max bytes in ioc rx packet
#define MAX_IOC_TX_PACKET_SIZE              32          // Max bytes in ioc rx packet
#define MAX_IOC_PACKET_SIZE                 MAX(MAX_IOC_RX_PACKET_SIZE,MAX_IOC_TX_PACKET_SIZE)
#define MAX_IOC_REFLASH_ATTEMPTS            5

#define LOW_BATTERY_WARNING_FREQUENCY       5   // Seconds between sending the same event
#define NO_WAN_BATTERY_WARNING_FREQUENCY    5   // Seconds between sending the same event
#define CRITICAL_BATTERY_WARNING_FREQUENCY  5   // Seconds between sending the same event

#define VOLTAGE_AVERAGE_SAMPLE_COUNT        20                              // Running average sample count
#define TEMPERATURE_AVERAGE_SAMPLE_COUNT    VOLTAGE_AVERAGE_SAMPLE_COUNT    // Running average sample count
#define AMPERAGE_AVERAGE_SAMPLE_COUNT       VOLTAGE_AVERAGE_SAMPLE_COUNT    // Running average sample count
#define MIN_SAMPLES_FOR_WATERMARK_AVERAGE   3   // Need at least three samples...

// Defines for the two different methods of voltage smoothing
typedef enum algs {
    VOLTAGE_ALGORITHM_1 = 1,
    VOLTAGE_ALGORITHM_2
}algs;


//
// Define SIMULATE_POWER_MODE_TRANSITION_FAILURE to add a
//  /proc/ioc/fail_next_transition entry which, when written a 1,
// will cause the IOC to purposely fail the next power transition request
// and return an error.  This should be disabled for ship.
#undef SIMULATE_POWER_MODE_TRANSITION_FAILURE

//
// Define ENABLE_PROC_IOC_RESET to add a /proc/ioc/reset entry which,
// when written a 1, will reset the IOC chip.  This should be disabled for ship.
// and return an error.
#undef ENABLE_PROC_IOC_RESET

//
// Define ENABLE_ALLOW_SLEEP_PROC to add a
//  /proc/ioc/allow_sleep entry which, when written a 0,
// will cause the IOC to purposely fail the next sleep request/
// and NOT go to sleep on request.  This should be disabled for ship.
#undef ENABLE_ALLOW_SLEEP_PROC

// Calculating board resistance is very unreliable.
// Define the following to use default values.
// Undef the following to calculate the resistance on-the-fly.
#define USE_DEFAULT_BOARD_RESISTANCE



//
// Type definitions
//
typedef struct IOC_ENABLE_REC {
    unsigned short  enable;          /* 1=ON, 0=OFF */
}IOC_ENABLE_REC, *IOC_ENABLE_REC_PTR;
    
typedef struct IOC_ENABLE_LOW_BATT_WARNINGS_REC {
    unsigned short  enable;          /* 1=ON, 0=OFF */
    int level;                       /* mvolt level of warning */
}IOC_ENABLE_LOW_BATT_WARNINGS_REC, *IOC_ENABLE_LOW_BATT_WARNINGS_REC_PTR;
    
typedef struct IOC_WAKE_CAUSE_REC  {
    int wake_source;
}IOC_WAKE_CAUSE_REC;

typedef struct IOC_FW_VERSION_REC {
    int version;
}IOC_FW_VERSION_REC;

typedef struct IOC_TWO_KEY_WAKE_REC {
    int wakeKey1;
    int wakeKey2;
} IOC_TWO_KEY_WAKE_REC; 
    
typedef struct IOC_BATTERY_INSTALLED_REC {
    unsigned int installed;
} IOC_BATTERY_INSTALLED_REC;

typedef struct IOC_BATTERY_VOLTAGE_REC {
    unsigned int mvolts;
} IOC_BATTERY_VOLTAGE_REC;

typedef struct IOC_BATTERY_TEMPERATURE_REC {
    signed int temperature;
}IOC_BATTERY_TEMPERATURE_REC;

typedef struct IOC_BATTERY_CURRENT_DRAW_REC {
    signed int draw;
}IOC_BATTERY_CURRENT_DRAW_REC; 

typedef struct IOC_BATTERY_CHARGE_STATUS_REC {
    unsigned int charging;
}IOC_BATTERY_CHARGE_STATUS_REC; 

typedef struct IOC_BATTERY_INFO_REC {
    unsigned int    mvolts;
    unsigned int    icon_mvolts;
    unsigned int    raw_mvolts;
    signed int      draw;
    signed int      temperature;
    unsigned int    charging;
    unsigned int    installed;             // 0: No Battery Installed,  1: Battery installed
}IOC_BATTERY_INFO_REC;

typedef IOC_BATTERY_INFO_REC Battery_Info_T;


#ifdef __KERNEL__
typedef struct packet_element {
    struct  list_head       list;
    unsigned char           packet[MAX_IOC_PACKET_SIZE];
    int                     length;
} packet_element_t;
#endif


//
// IOCTL Constants
//
#define IOC_PNLCD_SEQ_IOCTL                     _IOW(IOCTL_IOC_MAGIC_NUMBER, 0, PNLCD_SEQ_SEGMENTS_REC *)
#define IOC_PNLCD_VERT_SEQ_IOCTL                _IOW(IOCTL_IOC_MAGIC_NUMBER, 1, PNLCD_SEQ_SEGMENTS_REC *)
#define IOC_PNLCD_PAIRS_IOCTL                   _IOW(IOCTL_IOC_MAGIC_NUMBER, 2, PNLCD_SEGMENT_PAIRS_REC *)
#define IOC_PNLCD_VERT_PAIRS_IOCTL              _IOW(IOCTL_IOC_MAGIC_NUMBER, 3, PNLCD_SEGMENT_PAIRS_REC *)
#define IOC_PNLCD_MISC_IOCTL                    _IOW(IOCTL_IOC_MAGIC_NUMBER, 4, PNLCD_MISC_SEGMENTS_REC *)
#define IOC_ENABLE_IOC_IOCTL                    _IOW(IOCTL_IOC_MAGIC_NUMBER, 5, IOC_ENABLE_REC *)
#define IOC_PNLCD_ENABLE_IOCTL                  _IOW(IOCTL_IOC_MAGIC_NUMBER, 6, IOC_ENABLE_REC *)
#define IOC_KEYBOARD_ENABLE_IOCTL               _IOW(IOCTL_IOC_MAGIC_NUMBER, 7, IOC_ENABLE_REC *)
#define IOC_BATT_LV_WARN_ENABLE_IOCTL           _IOW(IOCTL_IOC_MAGIC_NUMBER, 8, IOC_ENABLE_LOW_BATT_WARNINGS_REC *)
#define IOC_BATT_CAPACITY_ENABLE_IOCTL          _IOW(IOCTL_IOC_MAGIC_NUMBER, 9, IOC_ENABLE_REC *)
#define IOC_QUERY_BATT_VOLTAGE_IOCTL            _IOR(IOCTL_IOC_MAGIC_NUMBER, 10, IOC_BATTERY_VOLTAGE_REC *)
#define IOC_QUERY_BATT_CAPACITY_IOCTL           _IOR(IOCTL_IOC_MAGIC_NUMBER, 11, IOC_BATTERY_CAPACITY_REC *)
#define IOC_QUERY_BATT_TEMP_IOCTL               _IOR(IOCTL_IOC_MAGIC_NUMBER, 12, IOC_BATTERY_TEMPERATURE_REC *)
#define IOC_QUERY_BATT_CHARGING_IOCTL           _IOR(IOCTL_IOC_MAGIC_NUMBER, 13, IOC_BATTERY_CHARGE_STATUS_REC *)
#define IOC_QUERY_BATT_CURRENT_DRAW_IOCTL       _IOR(IOCTL_IOC_MAGIC_NUMBER, 14, IOC_BATTERY_CURRENT_DRAW_REC *)
#define IOC_QUERY_BATT_INFO_IOCTL               _IOR(IOCTL_IOC_MAGIC_NUMBER, 15, IOC_BATTERY_INFO_REC *)
#define IOC_QUERY_BATT_INSTALLED_IOCTL          _IOR(IOCTL_IOC_MAGIC_NUMBER, 16, IOC_BATTERY_INSTALLED_REC *)
#define IOC_RESET_IOC_IOCTL                     _IOW(IOCTL_IOC_MAGIC_NUMBER, 17, NULL)
#define IOC_HOLD_LOCKOUT_IOCTL                  _IOW(IOCTL_IOC_MAGIC_NUMBER, 18, IOC_ENABLE_REC *)
#define IOC_QUERY_HOLD_LOCKOUT_IOCTL            _IOR(IOCTL_IOC_MAGIC_NUMBER, 19, IOC_ENABLE_REC *)
#define IOC_QUERY_FW_VERSION_IOCTL              _IOR(IOCTL_IOC_MAGIC_NUMBER, 20, IOC_FW_VERSION_REC *)
#define IOC_RESET_FIONA_IOCTL                   _IOW(IOCTL_IOC_MAGIC_NUMBER, 21, NULL)
#define IOC_SET_TWO_KEY_WAKE_IOCTL              _IOW(IOCTL_IOC_MAGIC_NUMBER, 22, IOC_TWO_KEY_WAKE_REC *)
#define IOC_TWO_KEY_WAKE_ENABLE_IOCTL           _IOW(IOCTL_IOC_MAGIC_NUMBER, 23, IOC_ENABLE_REC *)
#define IOC_ALL_KEY_WAKE_ENABLE_IOCTL           _IOW(IOCTL_IOC_MAGIC_NUMBER, 24, IOC_ENABLE_REC *)
#define IOC_QUERY_IOC_WAKE_CAUSE_IOCTL          _IOR(IOCTL_IOC_MAGIC_NUMBER, 25, IOC_WAKE_CAUSE_REC *)
#define IOC_IOCTLLast     25

#endif  // __IOC_H__
