/*
 * mainstone.c  --  SoC audio for Mainstone
 *
 * Copyright 2005 Wolfson Microelectronics PLC.
 * Author: Liam Girdwood
 *         liam.girdwood@wolfsonmicro.com or linux@wolfsonmicro.com
 *
 *  Mainstone audio amplifier code taken from arch/arm/mach-pxa/mainstone.c
 *  Copyright:	MontaVista Software Inc.
 * 
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  Revision history
 *    30th Oct 2005   Initial version.
 * 
 */
 
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/soc.h>
#include <sound/soc-dpm.h>

#include <asm/arch/pxa-regs.h>
#include <asm/arch/mainstone.h>
#include <asm/arch/audio.h>

#include "../codecs/wm8753.h"

static int mainstone_startup(snd_pcm_substream_t *substream)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_pcm_interface *pcm_i = pcm_c->pcm_i;
	
	if(pcm_i->type == SND_SOC_SSP && pcm_i->id == 1) {
		/* enable USB on the go MUX so we can use SSPFRM2 */
		MST_MSCWR2 |= MST_MSCWR2_USB_OTG_SEL;
		MST_MSCWR2 &= ~MST_MSCWR2_USB_OTG_RST;
	}
	return 0;
}

static void mainstone_shutdown(snd_pcm_substream_t *substream)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_pcm_interface *pcm_i = pcm_c->pcm_i;
			
	if(pcm_i->type == SND_SOC_SSP && pcm_i->id == 1) {
		/* disable USB on the go MUX so we can use ttyS0 */
		MST_MSCWR2 &= ~MST_MSCWR2_USB_OTG_SEL;
		MST_MSCWR2 |= MST_MSCWR2_USB_OTG_RST;
	}
}

static struct snd_soc_ops mainstone_ops = {
	.startup = mainstone_startup,
	.shutdown = mainstone_shutdown,
};

static long mst_audio_suspend_mask;

static int mainstone_suspend(struct snd_soc_machine *machine, pm_message_t state)
{
	mst_audio_suspend_mask = MST_MSCWR2;
	MST_MSCWR2 |= MST_MSCWR2_AC97_SPKROFF;
	return 0;
}

static int mainstone_resume(struct snd_soc_machine *machine)
{
	MST_MSCWR2 &= mst_audio_suspend_mask | ~MST_MSCWR2_AC97_SPKROFF;
	return 0;
}

static int mainstone_probe(struct snd_soc_machine *machine)
{
	MST_MSCWR2 &= ~MST_MSCWR2_AC97_SPKROFF;
	return 0;
}

static int mainstone_remove(struct snd_soc_machine *machine)
{
	MST_MSCWR2 |= MST_MSCWR2_AC97_SPKROFF;
	return 0;
}

/* example external dpm event callback */
int external_amp_event(struct snd_soc_codec *codec, int event)
{
	return 0;
}

/* example machine dpm widgets */
static const struct snd_soc_dpm_widget wm8753_dpm_widgets[] = {
SND_SOC_DPM_HP("Headphone Jack"),
SND_SOC_DPM_MIC("Mic1 Jack"),
SND_SOC_DPM_AMP("Ext Amp", external_amp_event),
};

/* example machine interconnections */
static const char* intercon[][3] = {
	
	/* headphone connected to LOUT1, ROUT1 */
	{"Headphone Jack", NULL, "LOUT1"},
	{"Headphone Jack", NULL, "ROUT1"},
	
	/* ext speaker connected to LOUT2, ROUT2 via amp */
	{"Ext Amp", NULL, "LOUT2"},
	{"Ext Amp", NULL, "ROUT2"},
	
	/* mic is connected to mic1 - with bias */
	{"MIC1", NULL, "Mic Bias"},
	{"MIC1N", NULL, "Mic Bias"},
	{"Mic Bias", NULL, "Mic1 Jack"},
	{"Mic Bias", NULL, "Mic1 Jack"},
	
	{"ACIN", NULL, "ACOP"},
	{NULL, NULL, NULL},
};

/* example unconnected codec pins */
static const struct snd_soc_dpm_pin wm8753_mainstone_pins[] = {
	{"RXP", 0},
	{"RXN", 0},
	{"MIC2", 0},
};

/* headphone detect support on my board */
static const char * hp_pol[] = {"Headphone", "Speaker"};
static const struct soc_enum wm8753_enum =
	SOC_ENUM_SINGLE(WM8753_OUTCTL, 1, 2, hp_pol);

static const snd_kcontrol_new_t wm8753_mainstone_controls[] = {
	SOC_SINGLE("Headphone Detect Switch", WM8753_OUTCTL, 6, 1, 0),
	SOC_ENUM("Headphone Detect Polarity", wm8753_enum),
};

/*
 * This is an example machine initialisation for a wm8753 connected to a 
 * Mainstone II. It is missing logic to detect hp/mic insertions and logic
 * to re-route the audio in such an event.
 */

static int mainstone_wm8753_init(struct snd_soc_codec *codec)
{
	int i, err;
	
	/* set up mainstone codec pins */
	for (i = 0; i < ARRAY_SIZE(wm8753_mainstone_pins); i++)
		snd_soc_dpm_set_connection(codec, &wm8753_mainstone_pins[i]);
	
	/* add mainstone specific controls */
	for (i = 0; i < ARRAY_SIZE(wm8753_mainstone_controls); i++) {
		if ((err = snd_ctl_add(codec->card, snd_soc_cnew(&wm8753_mainstone_controls[i],codec, NULL))) < 0)
			return err;
	}
	
	/* add mainstone specific widgets */
	for(i = 0; i < ARRAY_SIZE(wm8753_dpm_widgets); i++) {
		snd_soc_dpm_new_control(codec, &wm8753_dpm_widgets[i]);
	}
	
	/* set up mainstone specific audio path interconnects */
	for(i = 0; intercon[i][0] != NULL; i++) {
		snd_soc_dpm_connect_input(codec, intercon[i][0], intercon[i][1], intercon[i][2]);
	}
	snd_soc_dpm_sync(codec);
	
	return 0;
}

/* I've added all my different configs here atm for easier testing on ms II */
static struct snd_soc_machine_config codecs[] = {
{
	.name = "WM8753", 
	.sname = "WM8753 HiFi", 
	.interface = SND_SOC_I2S, 
	.int_num = 0,
	.init = mainstone_wm8753_init,
},
#if 0
	{.name = "WM8971", .sname = "WM8971", SND_SOC_I2S, 0},
	{.name = "WM8750", .sname = "WM8750", SND_SOC_I2S, 0},
	{.name = "WM8731", .sname = "WM8731", SND_SOC_I2S, 0},
	{.name = "WM8753", .sname = "WM8753 Voice", SND_SOC_SSP, 1},
	{.name = "AC97", .sname = "AC97 Audio", SND_SOC_AC97, 0},
#endif
};

static struct snd_soc_machine mainstone = {
	.name = "Mainstone",
	.probe = mainstone_probe,
	.remove = mainstone_remove,
	.suspend = mainstone_suspend,
	.resume = mainstone_resume,
	.ops = &mainstone_ops,
	.config = codecs,
	.nconfigs = ARRAY_SIZE(codecs),
};

static int __init snd_soc_init(void) 
{
	return snd_soc_register_machine(&mainstone);
}

static void __exit snd_soc_exit(void) 
{
	snd_soc_unregister_machine(&mainstone);
}

module_init(snd_soc_init); 
module_exit(snd_soc_exit);

/* Module information */ 
MODULE_AUTHOR("Liam Girdwood, liam.girdwood@wolfsonmicro.com, www.wolfsonmicro.com");
MODULE_DESCRIPTION("ALSA SoC Mainstone");
MODULE_LICENSE("GPL");
