/*
 * e-ink display framebuffer driver for bootloader
 *
 * Copyright (C) 2005-2006 Lab126, Inc.  All rights reserved.
 */

#include <common.h>
#include <command.h>

#ifdef CONFIG_EINK_APOLLO

#include "ub_boot_globals.h"
#include "ub_apollo.h"
#include "ub_eips.h"
#include "fb.h"

#include "images/image_manager.h"

#if PRAGMAS
    #pragma mark Externs, Local Definitions & Globals
    #pragma mark -
#endif

// Prototypes
//
extern void print_banner(char *banner_string, eips_cmd_type eips_cmd);
extern void print_usage(cmd_tbl_t *cmdtp, int argc, char *argv[]);
extern int  do_mmc_fatload(char *fat_load_addr, char *path_name);

// Definitions
//
#define FB_BASE             FRAMEBUFFER_START
#define FB_XRES             600
#define FB_YRES             800
#define FB_1BPP             1
#define FB_2BPP             2
#define FB_SIZE             ((FB_XRES * FB_YRES * FB_2BPP)/8)
#define FB_1BPP_WHITE       0xFF
#define FB_1BPP_BLACK       0x00
#define FB_2BPP_WHITE       0x00
#define FB_2BPP_BLACK       0xFF

#define FB_OFFSCREEN        CFG_MEMTEST_START
#define FB_OFF_SIZE         (CFG_MEMTEST_END - FB_OFFSCREEN)

#define FB_CMD_ARGS         7
#define FB_CMD_ARG          1

#define FB_CMD_DRAW_FILE    2
#define FB_CMD_DRAW_ARGC    3
#define FB_CMD_DRAW_X       3
#define FB_CMD_DRAW_Y       4
#define FB_CMD_DRAW_XY_ARGC 5

#define FB_CMD_ID_ARGC      2
#define FB_CMD_ID_VLR       2
#define FB_CMD_ID_VLR_ARGC  3

#define FB_ID_TEMPLATE_STR  "012_34_5678_"
#define FB_ID_SEPARATOR     '_'
#define FB_ID_FIELD_LLL     0   // "012"
#define FB_ID_FIELD_RR      1   // "34"
#define FB_ID_FIELD_PP      2   // "5678"

#ifndef CONFIG_USE_ALERT_PICTURES
#define ALERT               "ALERT\n"
#define ALERT_COL_START     5 

#define NO_BATTERY          "Please insert your battery.\n"
#define NO_BATTERY_STRS     1

#define LOW_BATTERY1        "Your battery is too low and Kindle is\n"
#define LOW_BATTERY2        "shutting down. You must recharge your\n"
#define LOW_BATTERY3        "Kindle battery to continue.\n"
#define LOW_BATTERY_STRS    3

#define POWER_CYCLE1        "You need to power cycle your Kindle.\n"
#define POWER_CYCLE2        "Please remove your battery, power adapter,\n"
#define POWER_CYCLE3        "and USB cable for 10 seconds before\n"
#define POWER_CYCLE4        "reinserting them.\n"
#define POWER_CYCLE_STRS    4

#define DOA1                "There is a problem with your Kindle.\n"
#define DOA2                "Please call Amazon Customer Service at\n"
#define DOA3                "866-321-8851 or 206-266-0927.\n"
#define DOA_STRS            3

#define MAX_ALERT_STRS      POWER_CYCLE_STRS
#endif

enum splash_picture_constraint
{
    splash_picture_top = 0,
    splash_picture_center,
    splash_picture_bottom,
    
    splash_picture_overlay,
    splash_picture_xy
};
typedef enum splash_picture_constraint splash_picture_constraint;

enum splash_picture_type
{
    splash_picture_logo = 0,
    
    splash_picture_service_menu_all,
    
    splash_picture_service_menu_update,
    splash_picture_service_menu_recover,
    splash_picture_service_menu_exit,
    
    splash_picture_update_snag,
    splash_picture_recover_snag,
    
    splash_picture_no_battery,
    splash_picture_low_battery,
    splash_picture_powercycle,
    splash_picture_doa,
    
    splash_picture_from_memory,
    
    splash_picture_unknown
};
typedef enum splash_picture_type splash_picture_type;

struct raw_image_type
{
    int xres,
        yres,
        bpp;
    
    u8  start[];
};
typedef struct raw_image_type raw_image_type;

enum fb_cmd_type
{
    fb_cmd_info,
    fb_cmd_overlay,
    fb_cmd_draw,
    fb_cmd_eips,
    fb_cmd_id,
    
    fb_num_cmds,
    fb_cmd_unknown
};
typedef enum fb_cmd_type fb_cmd_type;

// Globals
//
static bool g_init_done     = false;                  // Two-pass initialization.

static u8   *g_fb_base      = (u8 *)FB_BASE;          // Main <screen_info>:
static int  g_xres          = FB_XRES;                //    600 x
static int  g_yres          = FB_YRES;                //    800 @
static int  g_bpp           = FB_2BPP;                //    2bpp
static int  g_fb_size       = FB_SIZE;                //   

static u8   *g_offscreen    = (u8 *)FB_OFFSCREEN;     // Decompress images offscreen.
static int  g_off_size      = FB_OFF_SIZE;            //

static bool g_overlay = false;                        // Support for overlays.
static int  g_xstart;                                 // Support for byte-aligned...
static int  g_ystart;                                 // ...(x, y) coordinates.
static int  g_xend;                                   // Support for byte-aligned...
static int  g_yend;                                   // ...resolutions.

bool g_ack_timed_out = false;                         // If true, the controller can't be ack'd.

static char *fb_cmd_names[fb_num_cmds] =
{
    "info",                                           // fb_cmd_info
    "overlay",                                        // fb_cmd_overlay
    "draw",                                           // fb_cmd_draw
    "eips",                                           // fb_cmd_eips
    "id"                                              // fb_cmd_id
};

static u8 stretch_1bpp_to_2bbp_nybble_table[16] =
{
    0x00, 0x03, 0x0C, 0x0F, 0x30, 0x33, 0x3C, 0x3F, // 0x0..
    0xC0, 0xC3, 0xCC, 0xCF, 0xF0, 0xF3, 0xFC, 0xFF, // ..0xF
};

static const char fb_id_template_str[] = { FB_ID_TEMPLATE_STR };

#if PRAGMAS
    #pragma mark -
    #pragma mark E-ink Driver
    #pragma mark -
#endif

static void eink_fb_finish_init(void)
{
    // We only need to finish once.
    //
    if ( !g_init_done )
    {
        apollo_init(g_bpp);
        g_init_done = true;
    }
}

void eink_fb_get_screen_info(eink_screen_info *screen_info)
{
    if ( screen_info )
    {
        eink_fb_finish_init();
        
        screen_info->xres = g_xres;
        screen_info->yres = g_yres;
        screen_info->bpp  = g_bpp;
        screen_info->base = g_fb_base;
    }
}

unsigned long eink_get_bpp(void)
{
    eink_screen_info screen_info;
    
    eink_fb_get_screen_info(&screen_info);
    
    return ( (unsigned long)screen_info.bpp );
}

u8 *eink_get_framebuffer(void)
{
    eink_screen_info screen_info;
    
    eink_fb_get_screen_info(&screen_info);
    
    return ( screen_info.base );
}

void eink_fb_update_display(bool update_mode)
{
    eink_fb_finish_init();
    
    if ( g_overlay )
    {
        raw_image_type *image = (raw_image_type *)g_offscreen;
        int image_size = (image->xres * image->yres * FB_2BPP)/8;
        bool area_cleared, screen_clear = (bool)get_screen_clear();

        area_cleared = (bool)apollo_load_partial_data(image->start, image_size, g_xstart, g_ystart, g_xend, g_yend);
        
        if ( !(area_cleared && screen_clear) )
            apollo_display_partial();
        else
             update_mode = EINK_FULL_UPDATE_OFF;
        
        set_screen_clear(screen_clear && area_cleared);
        g_overlay = false;
    }
    else
    {
        bool screen_cleared = (bool)apollo_load_image_data(g_fb_base, g_fb_size) && (APOLLO_FX_NONE == apollo_get_display_fx());
        
        if ( !screen_cleared || (screen_cleared && !get_screen_clear()) )
        {
            if ( EINK_FULL_UPDATE_ON == update_mode )
                send_command(APOLLO_MANUAL_REFRESH_CMD);
                
            apollo_display();

            set_screen_clear(screen_cleared);
        }
    }
}

void eink_fb_clear_screen(void)
{
    eink_fb_finish_init();
    
    if ( !get_screen_clear() )
        memset(g_fb_base, FB_2BPP_WHITE, g_fb_size);
        
    eink_fb_update_display(EINK_FULL_UPDATE_ON);
}

void eink_fb_boot(void)
{
    set_fb_init_flag(BOOT_GLOBALS_FB_INIT_FLAG);
    
    apollo_set_init_display_speed(APOLLO_INIT_DISPLAY_FAST);
    send_command(APOLLO_SLEEP_MODE_CMD);
}

void eink_fb_remove(bool hard_reset)
{
    // If sending commands to the controller have timed out, we no longer want
    // to touch the controller, even if we're hard resetting.  Instead, we
    // simply want to tell ourselves that, on the next reboot, we need to
    // re-initialize the part from scratch (i.e., slowly).
    //
    if ( g_ack_timed_out )
        apollo_set_init_display_speed(APOLLO_INIT_DISPLAY_SLOW);
    else
    {
        // Otherwise, let's first get back to the default bit depth and clear the screen.
        //
        eink_fb_clear_screen();
        
        // From a hard reset, we'll only need to (re-)init the part.  Otherwise, we'll
        // need to wake it up first.
        //
        apollo_set_init_display_speed(hard_reset ? APOLLO_INIT_DISPLAY_FAST : APOLLO_INIT_DISPLAY_WAKE);
        
        // On hard resets, let's power all the way off.  Otherwise, just sleep.
        //
        if ( hard_reset )
        {
            send_command(APOLLO_STANDBY_MODE_CMD);
            udelay(APOLLO_STANDBY_CMD_DELAY);
            
            apollo_power_down(APOLLO_POWER_DOWN_FULL);
            set_fb_init_flag(0);
        }
        else
            send_command(APOLLO_SLEEP_MODE_CMD);
    }

    // Say we need to (re-)init now.
    //
    g_init_done = false;
}

void eink_fb_probe(void)
{
    g_init_done = false;
    
    if ( !get_screen_clear() )
        memset(g_fb_base, FB_2BPP_WHITE, g_fb_size);

    apollo_power_up();
}

#if PRAGMAS
    #pragma mark -
    #pragma mark Rendering
    #pragma mark -
#endif

static bool image_is_acceptable(int x_offset, int y_offset, int xres, int yres,
                                int image_xres, int image_yres)
{
    bool acceptable = true;
    
    // The limiting boundary must be within the screen's boundaries.
    //
    acceptable &= ((xres >= 0) && (xres <= g_xres));
    acceptable &= ((yres >= 0) && (yres <= g_yres));
    
    // Our starting and ending points must be byte aligned.
    //
    acceptable &= (0 == (x_offset % 8)) && (0 == (image_xres % 8));
   
    // The image must be within the limiting boundary.
    //
    acceptable &= (x_offset + image_xres) <= xres;
    acceptable &= (y_offset + image_yres) <= yres;

    // Print some diagnostics if the image isn't acceptable.
    //
    if ( !acceptable )
    {
        puts("Splash screen image couldn't be rendered:\n");
        printf(" Image screen bounds:   %3d x %3d\n", xres, yres);
        printf(" Image resolution:      %3d x %3d\n", image_xres, image_yres);
        printf(" Image offset:          %3d x %3d\n", x_offset, y_offset);
        printf(" Image row start mod 8: %3d\n", (x_offset % 8));
        printf(" Image width mod 8:     %3d\n", (image_xres % 8));
    }
    
    return ( acceptable );
}

static u8 get_image_byte(u8 *image, int bpp, int *byte)
{
    u8 image_byte;
    
    // At 2bpp, we're just byte per byte.
    //
    if ( FB_2BPP == bpp )
        image_byte = image[(*byte)++];
    else
    {
        // At 1bpp, we're two nybbles per byte.
        //
        int nybble = (*byte)++ >> 1;
        
        // Don't bother looking up FB_2BPP_WHITE in the table.
        //
        image_byte = image[nybble];
        
        // Otherwise, on odd bytes, we use the upper nybble; and on even
        // bytes we use the lower nybble.  Note:  We always start out
        // even (zero).
        //
        if ( FB_2BPP_WHITE != image_byte )
        {
            if ( 1 == (*byte & 1) )
                image_byte = stretch_1bpp_to_2bbp_nybble_table[(0xF0 & image_byte) >> 4];
            else
                image_byte = stretch_1bpp_to_2bbp_nybble_table[(0x0F & image_byte)];
        }
    }
    
    return ( image_byte );
}

static void load_splash_picture(splash_picture_type which_picture, splash_picture_constraint which_constraint)
{
    bool clear_background = true, image_in_memory = false, overlay = false;
    image_type which_image = image_unknown;
    int x_offset = 0, xres = g_xres,
        y_offset = 0, yres = g_yres;

    // First, translate <which_picture> into <which_image>.
    //
    switch ( which_picture )
    {
        case splash_picture_logo:
            which_image = image_picture_logo;
        break;
        
        case splash_picture_service_menu_all:
            which_image = image_picture_service_menu_all;
        break;
        
        case splash_picture_service_menu_update:
            which_image = image_picture_service_menu_update;
            clear_background = false;
        break;
        
        case splash_picture_service_menu_recover:
            which_image = image_picture_service_menu_recover;
            clear_background = false;
        break;
        
        case splash_picture_service_menu_exit:
            which_image = image_picture_service_menu_exit;
            clear_background = false;
        break;
        
        case splash_picture_update_snag:
            which_image = image_picture_update_snag;
        break;
                
        case splash_picture_recover_snag:
            which_image = image_picture_recover_snag;
        break;

#ifdef CONFIG_USE_ALERT_PICTURES
        case splash_picture_no_battery:
            which_image = image_picture_no_battery;
        break;
        
        case splash_picture_low_battery:
            which_image = image_picture_low_battery;
        break;

        case splash_picture_powercycle:
            which_image = image_picture_powercycle;
        break;
        
        case splash_picture_doa:
            which_image = image_picture_doa;
        break;
#endif

        case splash_picture_from_memory:
            image_in_memory = true;
        break;

        // Prevent compiler warnings.
        //
        default:
        break;
    }
    
    // Next, get the requested image into our "offscreen" buffer.
    //
    if ( image_in_memory || (image_unknown != which_image) )
    {
        raw_image_type *image = (raw_image_type *)g_offscreen;
        picture_info *image_info = NULL;
        u32 image_size = g_off_size;
        
        if ( image_in_memory || (SUCCESS == uncompress_image(which_image, (u8 *)image, &image_size, (void **)&image_info)) )
        {
            int image_xres, image_yres, image_bpp;
            u8 *image_start;
            
            if ( image_info->headerless )
            {
                image_xres  = image_info->xres;
                image_yres  = image_info->yres;
                image_bpp   = image_info->bpp;
                image_start = (u8 *)image;
            }
            else
            {
                image_xres  = image->xres;
                image_yres  = image->yres;
                image_bpp   = image->bpp;
                image_start = image->start;
            }
            
            // Now, constrain picture to specified portion of screen.
            //
            switch ( which_constraint )
            {
                // Constrain picture to top 1/3rd of screen.
                //
                case splash_picture_top:
                    y_offset = 0;
                    yres     = g_yres/3;
                break;
                
                // Constrain picture to bottom 2/3rds of screen and don't
                // clear the background.
                //
                case splash_picture_bottom:
                    y_offset = g_yres/3;
                    yres     = g_yres - y_offset;
                    
                    clear_background = false;
                break;
                
                // Start picture at specific (x, y) coordinate on screen, overlaying if
                // requested and possible.
                //
                case splash_picture_overlay:
                    overlay = (FB_2BPP == image_bpp) ? true : false;
                    
                    if ( !overlay )
                        clear_background = true;
                        
                    which_constraint = splash_picture_xy;
                
                case splash_picture_xy:
                    x_offset = image_in_memory ? g_xstart : image_info->x;
                    y_offset = image_in_memory ? g_ystart : image_info->y;
                    
                    // Fall back to centering image if either offset is
                    // negative.
                    //
                    if ( (0 > x_offset) || (0 > y_offset) )
                    {
                        which_constraint = splash_picture_center;
                        x_offset = y_offset = 0;
                    }
                 break;
                
                // Center picture.
                //
                case splash_picture_center:
                default:
                break;
            }
            
            // Ascertain the image's particulars and then ensure that they
            // are acceptable.
            //
            if ( image_is_acceptable(x_offset, y_offset, xres, yres, image_xres, image_yres) )
            {
                int x, x_start, x_end, rowbytes,
                    y, y_start, y_end, bytes;

                // Compute (x, y) start and end positions based on specified
                // constraints.
                //
                x_start = x_offset + ((splash_picture_xy == which_constraint) ? 0
                                   : ((xres - image_xres) >> 1));
                x_end   = x_start  + image_xres;
                
                y_start = y_offset + ((splash_picture_xy == which_constraint) ? 0
                                   : ((yres - image_yres) >> 1));                  
                y_end   = y_start  + image_yres;
                                
                // If we're overlaying, the image is already in the offscreen buffer.
                // Otherwise, we move it offscreen to onscreen.
                //
                if ( overlay )
                {
                    g_xstart = x_start; g_xend = x_end;
                    g_ystart = y_start; g_yend = y_end;
                    
                    g_overlay = true;
                }
                else
                {
                    // Make bpp-related adjustments.
                    //
                    x_start  = (x_start * FB_2BPP)/8;
                    x_end    = (x_end   * FB_2BPP)/8;
                    rowbytes = (g_xres  * FB_2BPP)/8;
    
                    // Perform blit after conditionally clearing specified background.
                    //
                    if ( clear_background )
                    {
                        u8 background = image_info->fill_white ? FB_2BPP_WHITE : FB_2BPP_BLACK;
                        memset(g_fb_base, background, g_fb_size);
                    }
                    
                    for (bytes = 0, y = y_start; y < y_end; y++)
                        for ( x = x_start; x < x_end; x++)
                            g_fb_base[(rowbytes * y) + x] = get_image_byte(image_start, image_bpp, &bytes);
                }
            }
        }
    }
}

void draw_splash_screen(splash_screen_type which_screen)
{
    bool splash = true, highlight = false;
#ifndef CONFIG_USE_ALERT_PICTURES
    char *alert_str[MAX_ALERT_STRS] = { 0 };
    int  num_alert_strs = 0;
    bool alert = false;
#endif
    
    // First, go render the specified splash screen offscreen in the
    // appropriate manner.
    //
    switch ( which_screen )
    {        
        // Logo-screen case.
        //
        case splash_screen_logo:
            load_splash_picture(splash_picture_logo, splash_picture_xy);
        break;
        
        // Service menu cases:  (1) Draw the service menu itself.
        //
        //
        case splash_screen_service_menu_all:
            load_splash_picture(splash_picture_service_menu_all, splash_picture_xy);
        break;
        
        //                      (2) Highlight the update menu item OR draw snag-key screen.
        //
        case splash_screen_update_snag:
            load_splash_picture(splash_picture_update_snag, splash_picture_xy);
        break;
        
        case splash_screen_service_menu_update:
            load_splash_picture(splash_picture_service_menu_update, splash_picture_overlay);
            highlight = true;
        break;
        
        //                      (3) Highlight the recover menu item OR draw snag-key screen.
        //
        case splash_screen_recover_snag:
            load_splash_picture(splash_picture_recover_snag, splash_picture_xy);
        break;
        
        case splash_screen_service_menu_recover:
            load_splash_picture(splash_picture_service_menu_recover, splash_picture_overlay);
            highlight = true;
        break;
        
        //                      (4) Highlight the exit menu item.
        //
        case splash_screen_service_menu_exit:
            load_splash_picture(splash_picture_service_menu_exit, splash_picture_overlay);
            highlight = true;
        break;

        // Diagnostic cases.
        //
        case splash_screen_no_battery:
#ifdef CONFIG_USE_ALERT_PICTURES
            load_splash_picture(splash_picture_no_battery, splash_picture_center);
#else
            num_alert_strs  = NO_BATTERY_STRS;
            alert_str[0]    = NO_BATTERY;

            alert           = true;
            splash          = false;
#endif
        break;
        
        case splash_screen_low_battery:
#ifdef CONFIG_USE_ALERT_PICTURES
            load_splash_picture(splash_picture_low_battery, splash_picture_center);
#else
            num_alert_strs  = LOW_BATTERY_STRS;
            alert_str[0]    = LOW_BATTERY1;
            alert_str[1]    = LOW_BATTERY2;
            alert_str[2]    = LOW_BATTERY3;
            
            alert           = true;
            splash          = false;
#endif
        break;
        
        case splash_screen_powercycle:
#ifdef CONFIG_USE_ALERT_PICTURES
            load_splash_picture(splash_picture_powercycle, splash_picture_center);
#else        
            num_alert_strs  = POWER_CYCLE_STRS;
            alert_str[0]    = POWER_CYCLE1;
            alert_str[1]    = POWER_CYCLE2;
            alert_str[2]    = POWER_CYCLE3;
            alert_str[3]    = POWER_CYCLE4;
        
            alert           = true;
            splash          = false;
#endif
        break;
        
        case splash_screen_doa:
#ifdef CONFIG_USE_ALERT_PICTURES
            load_splash_picture(splash_picture_doa, splash_picture_center);
#else        
            num_alert_strs  = DOA_STRS;
            alert_str[0]    = DOA1;
            alert_str[1]    = DOA2;
            alert_str[2]    = DOA3;
        
            alert           = true;
            splash          = false;
#endif
        break;
        
        // From-memory cases.
        //
        case splash_screen_from_memory_overlay:
            load_splash_picture(splash_picture_from_memory, splash_picture_overlay);
            splash = false;
        break;

        case splash_screen_from_memory_draw:
            load_splash_picture(splash_picture_from_memory, splash_picture_xy);
            splash = false;
        break;
                
        // Huh?
        //
        default:
            splash = false;
        break;
    }

    // Splash screen...
    //
    if ( splash )
    {
        bool update_mode = EINK_FULL_UPDATE_OFF;
        
        // If not highlighting, do a full refresh.
        //
        if ( !highlight )
            update_mode = EINK_FULL_UPDATE_ON;
        
        eink_fb_update_display(update_mode);
    }
    
#ifndef CONFIG_USE_ALERT_PICTURES
    // ...or alert.
    //
    if ( alert )
    {
        int i, col_start;
        
        // Print the general alert banner.
        //
        print_banner(ALERT, eips_cmd_buffer);
        
        // Set row and column starts.
        //
        if ( 1 == num_alert_strs )
            col_start = (EIPS_PUTS_COL_END - strlen(alert_str[0])) >> 1;
        else
            col_start = ALERT_COL_START;
        
        eips_set_row((EIPS_PUTS_ROW_END - num_alert_strs) >> 1);
        
        // Print the rest of the alert.
        //
        for ( i = 0; i < num_alert_strs; i++ )
            eips_puts(eips_cmd_screen, col_start, alert_str[i]);
    }
 #endif
}

#if PRAGMAS
    #pragma mark -
    #pragma mark U-Boot
    #pragma mark -
#endif

static int fb_info(void)
{
    eink_screen_info     screen_info;
    apollo_status_t      status;
    apollo_info_t        info;

    // Resolution and bpp information.
    //
    eink_fb_get_screen_info(&screen_info);

    printf("Screen resolution:  %d x %d\n",  screen_info.xres, screen_info.yres);
    printf("Bits per pixel:     %d\n",       screen_info.bpp);
    putc('\n');
    
    // Buffers and their sizes.
    //
    printf("Framebuffer base:   0x%08X (",   g_fb_base);   print_size(g_fb_size,  ")\n");
    printf("Offscreen buffer:   0x%08X (",   g_offscreen); print_size(g_off_size, ")\n");
    putc('\n');
    
    // Waveform, panel, temperature, and controller information.
    //
    apollo_get_status(&status);
    apollo_get_info(&info);
    
    printf(" Waveform version:  0x%02X\n", info.waveform_version);
    printf("       subversion:  0x%02X\n", info.waveform_subversion);
    printf("             type:  0x%02X\n", info.waveform_type);
    printf("          FPL lot:  0x%02X\n", info.fpl_lot_bcd);
    printf("         run type:  0x%02X\n", info.run_type);
    putc('\n');

    printf("     FPL platform:  0x%02X\n", info.fpl_platform);
    printf("              lot:  0x%04X\n", info.fpl_lot_bin);
    printf(" adhesive run no.:  0x%02X\n", info.adhesive_run_number);
    putc('\n');

    printf("   Version string:  %s\n",     apollo_get_waveform_version_string());
    putc('\n');

    printf("  Mfg data device:  0x%08X\n", info.mfg_data_device);
    printf("          display:  0x%08X\n", info.mfg_data_display);
    printf("       serial no.:  0x%08X\n", info.serial_number);
    printf("         checksum:  0x%08X\n", info.checksum);
    putc('\n');
        
    printf("  Min temperature:  %d\n",     info.waveform_temp_min);
    printf("  Max temperature:  %d\n",     info.waveform_temp_max);
    printf("            steps:  %d\n",     info.waveform_temp_inc);
    printf("             last:  %d\n",     info.last_temperature);
    putc('\n');

    printf("   Operation mode:  0x%02X\n", status.operation_mode);
    printf("    screen status:  0x%02X\n", status.screen);
    printf("     auto refresh:  0x%02X\n", status.auto_refresh);
    printf("              bpp:  0x%02X\n", status.bpp);
    putc('\n');

    printf("   Apollo version:  0x%02X\n", info.controller_version);
    putc('\n');
    
    return ( SUCCESS );
}

static int fb_draw(char *file, bool overlay, bool draw)
{
    char offscreen_addr[] = {"deadbeef"};
    int result = FAILURE;
    
    // <mmcinit ; fatload mmc 1 [g_offscreen] [file]>
    //
    sprintf(offscreen_addr, "%x", g_offscreen);
    
    if ( SUCCESS == do_mmc_fatload(offscreen_addr, file) )
    {
        draw_splash_screen(overlay ? splash_screen_from_memory_overlay : splash_screen_from_memory_draw);
        
        if ( draw )
            eink_fb_update_display(overlay ? EINK_FULL_UPDATE_OFF : EINK_FULL_UPDATE_ON);

        result = SUCCESS;
    }
    
    return ( result );
}

void overlay_bitmap(int xpos, int ypos, int mask, const char *filename)
{
    g_xstart = xpos, g_ystart = ypos;
    fb_draw((char *)filename, true, true);      // overlay,  draw
}

void paint_bitmap(int xpos, int ypos, const char *filename)
{
    g_xstart = xpos; g_ystart = ypos;
    fb_draw((char *)filename, false, false);    // don't overlay, don't draw
}

void paint_screen(int num)
{
    if ( FBIO_SCREEN_IN_RANGE(num) )
        draw_splash_screen((splash_screen_type)num);
}

void string_to_panel_id(char *string, panel_id_t *panel_id)
{
    if ( panel_id && string )
    {
        int i, count = 0, field = 0, length = strlen(fb_id_template_str);
        char temp_string[length];
        bool done = false;
        
        for ( i = 0; (i < length) && !done; i++ )
        {
            if ( FB_ID_SEPARATOR == fb_id_template_str[i] )
            {
                temp_string[count] = '\0';
                
                switch ( field++ )
                {
                    case FB_ID_FIELD_LLL:
                        panel_id->lot                 = (unsigned short)simple_strtoul(temp_string, NULL, 10);
                    break;
                    
                    case FB_ID_FIELD_RR:
                        panel_id->adhesive_run_number = (unsigned char)simple_strtoul(temp_string, NULL, 10);
                    break;
                
                    case FB_ID_FIELD_PP:
                        panel_id->platform            = apollo_get_fpl_platform_from_str(temp_string);
                        done = true;
                    break;
                }
                
                count = 0;
            }
            else
                temp_string[count++] = string[i];
        }
    }
}

void panel_id_to_string(panel_id_t *panel_id, char *string)
{
    if ( panel_id && string )
    {
        apollo_fpl_t fpl;

        fpl.platform            = panel_id->platform;
        fpl.lot                 = panel_id->lot;
        fpl.adhesive_run_number = panel_id->adhesive_run_number;
        
        apollo_get_panel_id_str_from_fpl(string, &fpl);
    }
}

int fb_id(panel_id_t *panel_id)
{
    if ( panel_id )
    {
        char panel_id_string[PANEL_ID_STR_SIZE];
        
        panel_id->override = (unsigned char)true;
        set_panel_id(panel_id);
        
        panel_id_to_string(panel_id, panel_id_string);
        setenv(PANEL_ID_ENV, panel_id_string);
        
        printf("%s=%s\n", PANEL_ID_ENV, panel_id_string);
    }
    else
    {
        clearenv(PANEL_ID_ENV);
        clear_panel_id();
    }

    return ( env_save() ? FAILURE : SUCCESS );
}

int fb_id_string(char *id)
{
    int result = FAILURE;
    
    if ( id )
    {
        panel_id_t panel_id;
        
        string_to_panel_id(id, &panel_id);
        result = fb_id(&panel_id);
    }
    
    return ( result );
}

int fb_id_init(void)
{
    char panel_id_string[PANEL_ID_STR_SIZE];
    apollo_fpl_t fpl;
    
    eink_fb_finish_init();
    
    apollo_get_fpl_version(&fpl);
    apollo_get_panel_id_str_from_fpl(panel_id_string, &fpl);
    
    return ( fb_id_string(panel_id_string) );
}

int do_fb(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
    fb_cmd_type i, fb_cmd = fb_cmd_unknown;
    bool overlay = false, draw = true;
    int result = FAILURE;
    
    for ( i = 0; i < fb_num_cmds && (fb_cmd_unknown == fb_cmd); i++ )
        if ( 0 == strcmp(argv[FB_CMD_ARG], fb_cmd_names[i]) )
            fb_cmd = i;
            
    switch ( fb_cmd )
    {
        // <fb info>
        //
        case fb_cmd_info:
            result = fb_info();
        break;
        
        // <fb overlay file [x y]>
        //
        case fb_cmd_overlay:
            overlay = true;
        goto draw_common;
        
        // <fb draw file [x y]>
        //
        case fb_cmd_draw:
        draw_common:
            if ( FB_CMD_DRAW_XY_ARGC == argc )
            {
                g_xstart = (int)simple_strtoul(argv[FB_CMD_DRAW_X], NULL, 10);
                g_ystart = (int)simple_strtoul(argv[FB_CMD_DRAW_Y], NULL, 10);
            }
            else
                g_xstart = g_ystart = FB_DRAW_CENTER;
            
            if ( FB_CMD_DRAW_ARGC <= argc )
                result = fb_draw(argv[FB_CMD_DRAW_FILE], overlay, draw);
            else
                goto usage;
        break;
        
        // <fb eips [-c|-n] [-b|-o|-s] [[xpos ypos string|filename]|num]>
        //
        case fb_cmd_eips:
            result = eips((argc-FB_CMD_ARG), &argv[FB_CMD_ARG]);
        break;
        
        // <fb id [LLL_RR_PP]>
        //
        case fb_cmd_id:
            if ( FB_CMD_ID_VLR_ARGC == argc )
                result = fb_id_string(argv[FB_CMD_ID_VLR]);
            else
            {
                if ( FB_CMD_ID_ARGC == argc )
                    result = fb_id(FB_CLEAR_PANEL_ID_OVERRIDE);
                else
                    goto usage;
            }
        break;
        
        // Huh?
        //
        default:
        usage:
            print_usage(cmdtp, argc, argv);
        break;
    }

    return ( result );
}

U_BOOT_CMD(
    fb, FB_CMD_ARGS, CMD_FLAG_NONE, do_fb,
    "fb      - framebuffer subsystem\n",
    "info\n"
    "    - prints miscellaneous information about the framebuffer out\n"
    "fb [draw | overlay] file [x y]\n"
    "    - fatloads 1bpp or 2bpp .raw file from MMC/SD card and draws it\n"
    "      at (x, y) or centered if (x, y) isn't specified, clearing the\n"
    "      background first\n"
    "    - or overlays the .raw file onto existing file, not clearing the\n"
    "      background\n"
    "fb eips [-c|-n] [-b|-o|-s] [[xpos ypos string|filename]|num]\n"
    "    - prints string to screen using 5x7 font at xpos ypos\n"
    "    - -c clears screen first and prints string if there\n"
    "    - -n sends string to buffer instead of screen\n"
    "    - -b draws filename as a bitmap at xpos ypos\n"
    "    - -o overlays filename as a bitmap at xpos ypos\n"
    "    - -s draws screen num\n"
    "fb id [LLL_RR_PP]\n"
    "    - sets the panel id to LLL_RR_PP, or clears the panel id if\n"
    "      LLL_RR_PP isn't specified (Lot_adhesive Run number_Platform)\n"
);

#endif // CONFIG_EINK_APOLLO
