/*
 * Copyright (C) 2006 Lab126, Inc.
 * Nick Vaccaro, nvaccaro@lab126.com
 * Mike Puckett, mikepuck@amazon.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Adapted by Nick Vaccaro <nvaccaro@lab126.com> from gunstix.c, 
 *  originally written by Craig Hughes <craig@gumstix.com>.
 *
 */

#include <config.h>
#include <common.h>
#include <fiona_ioc.h>
#include <asm/arch/pxa-regs.h>

#include "ub_boot_globals.h"
#include "fb.h"

#define SERIAL_NUMBER_DEFAULT       "AAAAAAAAAAAAAAAA"
#define MACH_TYPE_FIONA             627

#define FIONA_CALCULATE_OP_AMP_OFFSET_VERBOSE() fiona_calculate_op_amp_offset(1)
#define FIONA_CALCULATE_OP_AMP_OFFSET()         fiona_calculate_op_amp_offset(0)

#define OP_AMP_INIT_OVERRIDE()                  op_amp_init(true)
#define OP_AMP_INIT()                           op_amp_init(false)

#define GAIN_INIT_OVERRIDE()                    gain_init(true)
#define GAIN_INIT()                             gain_init(false)

#define RVE_INIT_OVERRIDE()                     rve_init(true)
#define RVE_INIT()                              rve_init(false)

/********************************************
 * Prototypes
 ********************************************/

extern int  onenand_get_serial_number(char *buf, int len);  /* cmd_nand.c */
extern bool serial_number_locked(void);

extern bool onenand_get_sync_default(void);
extern void onenand_sync_init(void);

extern bool onenand_sync_reads;

extern int fiona_calculate_op_amp_offset(int verbose);      /* cmd_fiona.c */

extern void console_env_overwrite(void);                    /* console.c */

extern void usb_force_sleep(void);                          /* isp1761.c */

unsigned int get_fiona_board_revision(bool verbose);

static bool save_env = false;

char dev_env[DEV_ENV_STR_SZ];

int board_init( void ){
  DECLARE_GLOBAL_DATA_PTR;

  gd->bd->bi_arch_number = MACH_TYPE_FIONA;	
  gd->bd->bi_boot_params = 0xa0000100;

  return 0;
}

int dram_init( void ){
  DECLARE_GLOBAL_DATA_PTR;

#if ( CONFIG_NR_DRAM_BANKS > 0 )
  gd->bd->bi_dram[0].start = PHYS_SDRAM_1;
  gd->bd->bi_dram[0].size  = PHYS_SDRAM_1_SIZE;
#endif
#if ( CONFIG_NR_DRAM_BANKS > 1 )  
  gd->bd->bi_dram[1].start = PHYS_SDRAM_2;
  gd->bd->bi_dram[1].size  = PHYS_SDRAM_2_SIZE;
#endif
#if ( CONFIG_NR_DRAM_BANKS > 2 )  
  gd->bd->bi_dram[2].start = PHYS_SDRAM_3;
  gd->bd->bi_dram[2].size  = PHYS_SDRAM_3_SIZE;
#endif
#if ( CONFIG_NR_DRAM_BANKS > 3 )
  gd->bd->bi_dram[3].start = PHYS_SDRAM_4;
  gd->bd->bi_dram[3].size  = PHYS_SDRAM_4_SIZE;
#endif
	
  return 0;
}

#ifdef CONFIG_SHOW_BOOT_PROGRESS

static int progress_location = UPDATE_PROGRESS_START;
static int refresh_location  = 0;

extern void progress_bar_error(int location);
extern void progress_bar(int location);
extern void progress_bar_init(void);

static int pin_progress(int location)
{
	int pinned_location = location;
	
	// Pin <location> to <MIN_PROGRESS>..<MAX_PROGRESS>.
	//
	if ( MIN_PROGRESS > location )
		pinned_location = MIN_PROGRESS;
	else
		if ( MAX_PROGRESS < location )
			pinned_location = MAX_PROGRESS;

	return ( pinned_location );
}

void set_progress_bar(int location)
{
	bool error = (0 > location) ? true : false;
	
	// Pin <location> to approprate range.
	//
	progress_location = pin_progress(error ? -location : location);
	
	// Update progress bar with new location.
	//
	if ( error )
	{
		printf("\n\nERROR: Boot location %d\n", location);
		progress_bar_error(progress_location);
	}
	else
	{
		//printf("boot progress = %d\n", progress_location);
		progress_bar(progress_location);
	}
}

void update_progress_bar(void)
{
	set_progress_bar(progress_location + 1);
	refresh_location = progress_location;
}

void refresh_progress_bar(void)
{
    if ( refresh_location != progress_location )
    {
        progress_bar_init();
        set_progress_bar(refresh_location);
    }
}

int get_remaining_progress(void)
{
	int remaining_progress = MAX_PROGRESS - progress_location;
	
	// If we're at max, refresh us back to the start.
	//
	if ( 0 == remaining_progress )
	{
        refresh_progress_bar();
        remaining_progress = MAX_PROGRESS - progress_location;
	}
	
	return ( remaining_progress );
}

void show_boot_progress(int location)
{
    // For <bootm> progress, peg us to SHOW_BOOT_PROGRESS_MAX.  Also,
    // don't bother setting the progress bar if we're already at the
    // the specified <location>.
    //
    if ( (SHOW_BOOT_PROGRESS_MAX > location) && (progress_location < location) )
        update_progress_bar();
    else
    {
        // If <bootm> is erroring out, show it.
        //
        if ( 0 > location )
            set_progress_bar(location);
    }
}

#endif // CONFIG_SHOW_BOOT_PROGRESS

unsigned int
get_fiona_board_revision_str(char *str, bool verbose)
{
    unsigned int rev = 0;
    const char *BOARD_REV_STRINGS[MAX_SUPPORTED_BOARD_REVS] = \
        { "P2","P3","P3.5","EVT1","EVT2","EVT3","REV1","REV2","REV3" };

    str[0] = '\0';

    // Read the board ID register to determine revision #.
    //
    rev = get_fiona_board_revision(verbose);
    
    if ( (rev > FIONA_BOARD_REV3) || 
         ((rev-FIONA_BOARD_P2) > MAX_SUPPORTED_BOARD_REVS) )
    {
        printf("Boards newer than REV3 are not yet supported!\n");
    }
    else
    {        
        // Subtract FIONA_BOARD_P2 to make it 0-based.
        //
        strcpy(str,BOARD_REV_STRINGS[rev-FIONA_BOARD_P2]);
    }

    return ( rev );
}



unsigned int
get_fiona_board_revision(bool verbose)
{
    unsigned int board_rev = *((unsigned int *)PHYS_BOARD_REV_ID);
    unsigned int rev = 0;
    char         *errStr = NULL;

    board_rev &= BOARD_REV_ID_MASK;

    switch (board_rev)  {
        // P1 and P2 boards are the same REV number...
        case BOARD_REV_ID_P2:
            rev = FIONA_BOARD_P2;
            break;

        case BOARD_REV_ID_P3:
            rev = FIONA_BOARD_P3;
            break;

        case BOARD_REV_ID_P3_5:
            rev = FIONA_BOARD_P3_5;
            break;

        case BOARD_REV_ID_EVT1:
            rev = FIONA_BOARD_EVT1;
            break;

        case BOARD_REV_ID_EVT2:
            rev = FIONA_BOARD_EVT2;
            break;

        case BOARD_REV_ID_EVT3:
            rev = FIONA_BOARD_EVT3;
            break;

        case BOARD_REV_ID_DVT1:
            rev = FIONA_BOARD_DVT1;
            break;

        case BOARD_REV_ID_REV3:
            printf("Saw a REV3 board, spoofing as a REV2...\n");
            // Fall thru to REV2 code

        case BOARD_REV_ID_REV2:
        	rev = FIONA_BOARD_REV2;
        	break;

        default:
            if ( verbose )
                printf("Unknown Fiona board revision ID 0x%x !\n",rev);
            rev = 0;
            break;
    }

#if defined(BUILD_FOR_REV2_HARDWARE)
    if ((board_rev != BOARD_REV_ID_DVT1) && (board_rev != BOARD_REV_ID_REV2) && (board_rev != BOARD_REV_ID_REV3))  {
        errStr = "REV2";
        rev = FIONA_BOARD_REV2;
        goto errExit;
    }
#elif defined(BUILD_FOR_DVT1_HARDWARE)
    if (board_rev != BOARD_REV_ID_DVT1)  {
        errStr = "DVT1";
        rev = FIONA_BOARD_DVT1;
        goto errExit;
    }
#elif defined(BUILD_FOR_EVT3_HARDWARE)
    if (board_rev != BOARD_REV_ID_EVT3)  {
        errStr = "EVT3";
        rev = FIONA_BOARD_EVT3;
        goto errExit;
    }
#elif defined(BUILD_FOR_EVT2_HARDWARE)
    if (board_rev != BOARD_REV_ID_EVT2)  {
        errStr = "EVT2";
        rev = FIONA_BOARD_EVT2;
        goto errExit;
    }
#elif defined(BUILD_FOR_EVT1_HARDWARE)
    if ((board_rev < BOARD_REV_ID_EVT1) || (board_rev == BOARD_REV_ID_P3_5))  {
        errStr = "EVT1";
        rev = FIONA_BOARD_EVT1;     // Bug on EVT1 boards - id incorrect
                                    // If we're built for EVT1, return EVT1 ID
        goto errExit;
    }
#elif defined(BUILD_FOR_P35_HARDWARE)
    if (board_rev != BOARD_REV_ID_P3_5)  {
        errStr = "P3.5";
        rev = FIONA_BOARD_P3_5;
        goto errExit;
    }
#elif defined(BUILD_FOR_P3_HARDWARE)
    if (board_rev != BOARD_REV_ID_P3)  {
        errStr = "P3";
        rev = 0;
        goto errExit;
    }
#endif
    goto exit;

errExit:
    if ( verbose )
    {
        printf("\nERROR !! This bootloader supports %s boards only.\n",errStr);
        printf("Assuming invalid board ID %d and treating this board as a %s.\n",(board_rev>>BOARD_REV_ID_SHIFT),errStr);
    }
    
exit:
    /* Read the board ID register to determine revision # */
    return(rev);
}

#ifdef CONFIG_REVISION_TAG
void setup_revision_tag(struct tag **paramsHndl)
{
    (*paramsHndl)->hdr.tag = ATAG_REVISION;
    (*paramsHndl)->hdr.size = tag_size (tag_revision);
    (*paramsHndl)->u.revision.rev = get_fiona_board_revision(FIONA_VERBOSE_REVISION);
	(*paramsHndl) = tag_next ((*paramsHndl));
}
#endif /* CONFIG_REVISION_TAG */

#ifdef CONFIG_BOARD_RESISTANCE_TAG
unsigned long fiona_calculate_board_resistance(int verbose);
void setup_board_resistance_tag(struct tag **paramsHndl)
{
    unsigned long resistance = fiona_calculate_board_resistance(0);

    (*paramsHndl)->hdr.tag = ATAG_BOARD_RESISTANCE;
    (*paramsHndl)->hdr.size = tag_size (tag_board_resistance);
    (*paramsHndl)->u.board_resistance.mohms = resistance;
	(*paramsHndl) = tag_next ((*paramsHndl));
}
#endif /* CONFIG_BOARD_RESISTANCE_TAG */

static bool get_serial_number(char *serial, int serial_len)
{
    bool serial_number_in_globals = false;
    char *serial_num;
    
    if ( NULL == (serial_num = get_serial_num()) )
    {
        onenand_get_serial_number(serial, serial_len);
        
        if (0 == serial[0])
            strcpy(serial, SERIAL_NUMBER_DEFAULT);
            
        set_serial_num(serial);
    }
    else
    {
        serial_number_in_globals = true;
        strcpy(serial, serial_num);
    }
        
    return ( serial_number_in_globals );
}

void serial_number_init(void)
{
    char serial[SERIAL_NUM_SIZE];

    // If the serial number is in the globals already, we'll
    // need to print some info to match what happens when we
    // must read it out of OTP and put it into the globals.
    //
    if ( get_serial_number(serial, SERIAL_NUM_SIZE) )
        printf("serial#=%s %s\n", serial, (serial_number_locked() ? "(locked)" : ""));
}

void setup_serial_tag (struct tag **params)
{
    char serial[SERIAL_NUM_SIZE];

    get_serial_number(serial, SERIAL_NUM_SIZE);

	(*params)->hdr.tag = ATAG_SERIAL;
	(*params)->hdr.size = tag_size (tag_serialnr);

	memcpy((*params)->u.serialnr.data, serial, sizeof(serial));

	(*params) = tag_next ((*params));
}

static bool env_vers_init(void)
{
    char env_vers_str[ENV_VERS_STR_SZ];
    bool env_vers_state = ENV_VERS_BAD;
    
    // If the environment's version doesn't exist, our environment is
    // bad.
    //
    if ( getenv_r(ENV_VERS, env_vers_str, ENV_VERS_STR_SZ) > 0 )
    {
        unsigned long env_vers = simple_strtoul(env_vers_str, NULL, 10);

        // Our environment is good when the environment's version exists
        // and it matches what we think it should be.
        //
        if ( CONFIG_ENV_VERS == env_vers )
            env_vers_state = ENV_VERS_GOOD;
    }        
    
    return ( env_vers_state );
}

static void onenand_sync_reads_init(void)
{
	bool saved_onenand_sync_reads = onenand_sync_reads;
	char sync_rd[ONENAND_SYNC_RD_STR_SZ];
	sync_rd[0] = '\0';
	
	if ( getenv_r(ONENAND_SYNC_RD_ENV, sync_rd, ONENAND_SYNC_RD_STR_SZ) > 0 )
	{
		if ( 0 == strcmp(sync_rd, "false") )
			onenand_sync_reads = false;
		else
		{
			if ( 0 == strcmp(sync_rd, "true") )
				onenand_sync_reads = true;
			else
			    sync_rd[0] = '\0';
		}
	}

	if ( '\0' == sync_rd[0] )
	{
		onenand_sync_reads = (get_fiona_board_revision(FIONA_QUIET_REVISION) >= FIONA_BOARD_REV2)
			? onenand_get_sync_default()
			: false;

		strcpy(sync_rd, (onenand_sync_reads ? "true" : "false"));
		setenv(ONENAND_SYNC_RD_ENV, sync_rd);
	}
	
	if ( saved_onenand_sync_reads != onenand_sync_reads)
        onenand_sync_init();
}

static void hw_flags_init(void)
{
    char hw_flags_string[HW_FLAGS_ENV_STR_SZ];
    unsigned long hw_flags;
    
    if ( getenv_r(HW_FLAGS_ENV, hw_flags_string, HW_FLAGS_ENV_STR_SZ) > 0 )
    {
        hw_flags = simple_strtoul(hw_flags_string, NULL, 16);
        set_hw_flags((hw_flags_t *)&hw_flags);
    }
    else
    {
        hw_flags = *(unsigned long *)get_hw_flags();
        sprintf(hw_flags_string, "%08X", hw_flags);
        
        setenv(HW_FLAGS_ENV, hw_flags_string);
    }
}

static bool bat_level_env_valid(void)
{
    bool bat_level_env_valid_status = false;
    
    if ( getenv(BAT_LEVEL_0_ENV) && getenv(BAT_LEVEL_1_ENV) && getenv(BAT_LEVEL_2_ENV) && getenv(BAT_LEVEL_3_ENV) )
        bat_level_env_valid_status = true;
       
    return ( bat_level_env_valid_status );
}

static void bat_level_init(void)
{
    char battery_watermarks_string[BAT_LEVEL_ENV_STR_SZ];
    bat_level_t battery_watermarks;
    
    // Get the battery watermarks from the globals.
    //
    memcpy(&battery_watermarks, get_battery_watermarks(), sizeof(bat_level_t));
    
    // If we're not supposed to update them, initialize the watermarks
    // to either the built-in values or, if available, from the
    // environment.
    //
    if ( (0 == battery_watermarks.update_env) || !bat_level_env_valid() )
    {
        battery_watermarks.stage[BAT_LEVEL_UPDT] = FIONA_WATERMARK_UPDT;
        battery_watermarks.stage[BAT_LEVEL_WAN]  = FIONA_WATERMARK_WAN;
        battery_watermarks.stage[BAT_LEVEL_LOW]  = FIONA_WATERMARK_LOW;
        battery_watermarks.stage[BAT_LEVEL_CRIT] = FIONA_WATERMARK_CRIT;

        // If any of the watermarks is missing from the environment, say that
        // we need to update them.
        //
        if ( getenv_r(BAT_LEVEL_0_ENV, battery_watermarks_string, BAT_LEVEL_ENV_STR_SZ) > 0 )
            battery_watermarks.stage[BAT_LEVEL_UPDT] = simple_strtoul(battery_watermarks_string, NULL, 10);
        else
            battery_watermarks.update_env = 1;

        if ( getenv_r(BAT_LEVEL_1_ENV, battery_watermarks_string, BAT_LEVEL_ENV_STR_SZ) > 0 )
            battery_watermarks.stage[BAT_LEVEL_WAN] = simple_strtoul(battery_watermarks_string, NULL, 10);
        else
            battery_watermarks.update_env = 1;
        
        if ( getenv_r(BAT_LEVEL_2_ENV, battery_watermarks_string, BAT_LEVEL_ENV_STR_SZ) > 0 )
            battery_watermarks.stage[BAT_LEVEL_LOW] = simple_strtoul(battery_watermarks_string, NULL, 10);
        else
            battery_watermarks.update_env = 1;
        
        if ( getenv_r(BAT_LEVEL_3_ENV, battery_watermarks_string, BAT_LEVEL_ENV_STR_SZ) > 0 )
            battery_watermarks.stage[BAT_LEVEL_CRIT] = simple_strtoul(battery_watermarks_string, NULL, 10);
        else
            battery_watermarks.update_env = 1;
        
        // Update the globals.
        //
        set_battery_watermarks(&battery_watermarks);
    }

    // If we're supposed to update the battery watermarks, then
    // do so now.
    //
    if ( 1 == battery_watermarks.update_env )
    {
        // Save each of the battery watermarks out to the environment.
        //
        sprintf(battery_watermarks_string, "%d", battery_watermarks.stage[BAT_LEVEL_UPDT]);
        setenv(BAT_LEVEL_0_ENV, battery_watermarks_string);
        
        sprintf(battery_watermarks_string, "%d", battery_watermarks.stage[BAT_LEVEL_WAN]);
        setenv(BAT_LEVEL_1_ENV, battery_watermarks_string);
        
        sprintf(battery_watermarks_string, "%d", battery_watermarks.stage[BAT_LEVEL_LOW]);
        setenv(BAT_LEVEL_2_ENV, battery_watermarks_string);
        
        sprintf(battery_watermarks_string, "%d", battery_watermarks.stage[BAT_LEVEL_CRIT]);
        setenv(BAT_LEVEL_3_ENV, battery_watermarks_string);
        
        // Say that we've update the environment and update the globals
        // to reflect that as well.
        //
        battery_watermarks.update_env = 0;
        set_battery_watermarks(&battery_watermarks);
        save_env = true;
    }
}

void op_amp_init(bool override)
{
    char op_amp_string[ENV_OP_AMP_STR_SZ];

#ifdef CONFIG_CALCULATE_OP_AMP_OFFSET
    if ( !override )  {
        int offset = FIONA_CALCULATE_OP_AMP_OFFSET() 
        if (offset != get_op_amp_offset()) {
            set_op_amp_offset(offset); 
            set_update_op_amp_offset_env(1);
        }
    }
#endif
    
    op_amp_string[0] = '\0';
    
    // If we're not supposed to update the op-amp offset value, (re)initialize
    // it from the environment if it's available.
    //
    if ( 0 == get_update_op_amp_offset_env() )
    {
        // If the op-amp offset value is missing from the environment, say
        // that we need to update it.  Otherwise, convert it from a string
        // to a number.
        //
        if ( getenv_r(ENV_OP_AMP, op_amp_string, ENV_OP_AMP_STR_SZ) > 0 )
        {
            bool negative = ('-' == op_amp_string[0]) ? true : false;
            int op_amp_offset = (int)simple_strtoul(&op_amp_string[negative], NULL, 10);

            set_op_amp_offset(negative ? -op_amp_offset : op_amp_offset);
        }
        else
            set_update_op_amp_offset_env(1);
    }
    
    // If we're supposed to update the op-amp offset value, then do so now.
    //
    if ( 1 == get_update_op_amp_offset_env() )
    {
        // Save the op-amp value out to the environment.
        //
        sprintf(op_amp_string, "%d", get_op_amp_offset());
        setenv(ENV_OP_AMP, op_amp_string);
        
        // Save thay we've updated environment.
        //
        set_update_op_amp_offset_env(0);
        save_env = true;
    }
    
    if ( strlen(op_amp_string) )
        printf("%s=%s\n", ENV_OP_AMP, op_amp_string);
}

void gain_init(bool override)
{
    char gain_string[ENV_GAIN_STR_SZ];
    gain_string[0] = '\0';
    
    // If the gain value is available from the environment, use it to set the boot globals with.
    // Otherwise, use the boot globals to initialize the environment.
    //
    if ( !override && (getenv_r(ENV_GAIN, gain_string, ENV_GAIN_STR_SZ) > 0) )
    {
        unsigned long gain_value = simple_strtoul(gain_string, NULL, 10);
        set_gain_value(gain_value);
    }
    else
    {
        sprintf(gain_string, "%d", get_gain_value());
        setenv(ENV_GAIN, gain_string);
        save_env = true;
    }
    
    if ( strlen(gain_string) )
        printf("%s=%s\n", ENV_GAIN, gain_string);
}

void rve_init(bool override)
{
    char rve_string[ENV_RVE_STR_SZ];
    rve_string[0] = '\0';
    
    // If the reference voltage error value is available from the environment,
    // use it to set the boot globals with.  Otherwise, use the boot globals
    // to initialize the environment.
    //
    if ( !override && (getenv_r(ENV_RVE, rve_string, ENV_RVE_STR_SZ) > 0) )
    {
        unsigned long rve_value = simple_strtoul(rve_string, NULL, 10);
        set_rve_value(rve_value);
    }
    else
    {
        sprintf(rve_string, "%d", get_rve_value());
        setenv(ENV_RVE, rve_string);
        save_env = true;
    }
    
    if ( strlen(rve_string) )
        printf("%s=%s\n", ENV_RVE, rve_string);
}

static bool panel_id_init(void)
{
    char panel_id_string[PANEL_ID_STR_SIZE];
    bool env_saved = false;
    
    if ( getenv_r(PANEL_ID_ENV, panel_id_string, PANEL_ID_STR_SIZE) > 0 )
    {
        panel_id_t *panel_id = get_panel_id();
        
        if ( !panel_id->override )
        {
            string_to_panel_id(panel_id_string, panel_id);
            env_saved = SUCCESS == fb_id(panel_id);
        }
            
        printf("%s=%s\n", PANEL_ID_ENV, panel_id_string);
    }
    else
        env_saved = SUCCESS == fb_id_init();
        
    return ( env_saved );
}

void env_restore_r(void)
{
    // Save the panel ID.
    //
    panel_id_t *panel_id = get_panel_id();

    // Restore the environment to its default state.
    //
    env_restore();
    
    // Restore the console environment.
    //
    console_env_overwrite();

    // Restore onenand_sync_reads to its default.
    //
    onenand_sync_reads_init();
    
    // Restore the hardware flags.
    //
    hw_flags_init();
    
    // Restore the battery levels.
    //
    bat_level_init();
    
    // Restore op-amp offset value.
    //
    OP_AMP_INIT();
    
    // Restore gain value.
    //
    GAIN_INIT();
    
    // Restore reference voltage error value.
    //
    RVE_INIT();
    
    // Clear the dev mode.
    //
    set_dev_mode(0);

    // Restore the panel ID to the environment if
    // it's been overridden, otherwise clear it.
    //
    // Note:  Calling fb_id() saves the environment
    //        variables.
    //
    if ( panel_id->override )
        fb_id(panel_id);
    else
        fb_id(FB_CLEAR_PANEL_ID_OVERRIDE);
}

static void env_init_r(void)
{
    // ENV: env_vers.
    //
    if ( ENV_VERS_BAD == env_vers_init() )
        env_restore_r();
    else
    {
        // ENV: onenand_sync_reads.
        //
        onenand_sync_reads_init();
                
        // ENV: hw_flags.
        //
        hw_flags_init();
        
        // ENV: bat_level#.
        //
        bat_level_init();
        
        // ENV: op_amp#.
        //
        OP_AMP_INIT();
        
        // ENV: gain#.
        //
        GAIN_INIT();
        
        // ENV: rve#.
        //
        RVE_INIT();
        
        // ENV: dev.
        //
        set_dev_mode(0);
        
        if ( getenv_r(DEV_ENV, dev_env, DEV_ENV_STR_SZ) > 0 )
            set_dev_mode(1);

        // ENV: panel#.
        // 
        // Note:  If we actually need to initialize the panel ID
        //        environment variable, we will have saved the
        //        entire environment, so there's no need to do it
        //        again.
        //
        if ( panel_id_init() )
            save_env = false;
        
        // Save the environment if necessary.
        //
        if ( save_env )
            env_save();
     }
}

int misc_init_r(void)
{
    // Perform environment-related initializations.
    //
    env_init_r();
    
    // Get the device's serial number.
    //
    serial_number_init();

	// Put USB into a low-power state.
    //
	usb_force_sleep();

    // All done.
    //
    return ( SUCCESS );
}

/* vim: set ts=4 sw=4 tw=80: */

