#include <config.h>
#include <version.h>
#include <asm/arch/pxa-regs.h>

#ifndef CONFIG_FIONA

/*
 * Local flags to enable/disable features
 */

/* Define INCLUDE_LED_BLINKER to include blink_led call for blinking heartbeat LED */
#undef INCLUDE_LED_BLINKER

/* Define CONFIG_WAKE_ON_SLEEP_ONLY to wake ONLY if sleep was only cause of "reset".
 * i.e. If a watchdog reset also occurred, don't resume from sleep..
 *
 * If it's NOT defined, we'll wake any time we're told we went to sleep, irrespective
 * of other possible reset conditions. -njv
 */
#undef CONFIG_WAKE_ON_SLEEP_ONLY

.macro CPWAIT
	mrc	 p15,0,r0,c2,c0,0
	mov	 r0,r0
	sub	 pc,pc,#4
.endm

// Define the following positions to make driving JTAG easier
//
.globl memsetup
.globl mem_init_complete


//
// Start of Memory Setup Code
//
.globl start_mem_controller_init
memsetup:

    /* 
     * If we're the bootloader running from our blessed location in RAM,
     * then we head back as memory is already setup.
     */
    ldr r0, =BLESSED_RAM_LOCATION   /* */
	cmp	pc, r0                      /** test if we're in SDRAM at blessed location	**/
	bhs	end_of_memsetup			    /** if we are, then jump		**/

    //
    // JTAG Support
    //
    // To have this bootloader behave as a JTAG-specific bootloader, all
    // that needs to be done is to load the bootloader image via JTAG to
    // a memory address that IS NOT >= BLESSED_RAM_LOCATION.  This code will check
    // the address it's running from, and if it's not the actual blessed bootloader
    // running from the blessed location in RAM, it will then check to see if
    // we're running from RAM.  If we are running from RAM, but not at the blessed
    // location, then we must want JTAG support and we'll skip the appropriate stuff
    // for doing JTAG recovery.
    //

    // We've determined we're not the bootloader running out of RAM at the blessed location.
    // See if we're running from RAM.  If we are, this must be a JTAG request.
    //
	cmp	pc, #CFG_DRAM_BASE      /** test if we're in SDRAM		**/
	bhs	init_gpios              /** if we are, then this is a JTAG recovery request, so
                                    jump wake code and go init gpio's	**/

#if defined(CONFIG_FIONA) && defined(CONFIG_POWER_SAVINGS)

    /* Check if we woke from a GPIO1 event */
    ldr     r0, =PEDR
    ldr     r0, [r0]
    tst     r0, #1
    beq     init_gpios  /* Branch if we DIDN'T wake from sleep */

    /*
     * If we just woke up (came out of sleep mode), we
     * won't reinitialize the GPIOs.
     */
    ldr     r0, =RCSR
    ldr     r1, [r0]

#ifdef CONFIG_WAKE_ON_SLEEP_ONLY
/*  If we want ONLY Sleep mode, do the following...  */
    cmp     r1, #4
    beq     no_init_gpios   /* Branch if we DID wake from sleep */
#else
    tst     r1, #4          /* Check Sleep Mode bit */
    bne     no_init_gpios   /* Branch if we DID wake from sleep */
#endif

init_gpios:

#endif /* CONFIG_FIONA */

	/* First, setup GPIOs */

	ldr		r0,	=GPSR0
	ldr		r1,	=CFG_GPSR0_VAL
	str		r1,   [r0]

	ldr		r0,	=GPSR1
	ldr		r1,	=CFG_GPSR1_VAL
	str		r1,   [r0]

	ldr		r0,	=GPSR2
	ldr		r1,	=CFG_GPSR2_VAL
	str		r1,   [r0]

	ldr		r0,	=GPCR0
	ldr		r1,	=CFG_GPCR0_VAL
	str		r1,   [r0]

	ldr		r0,	=GPCR1
	ldr		r1,	=CFG_GPCR1_VAL
	str		r1,   [r0]

	ldr		r0,	=GPCR2
	ldr		r1,	=CFG_GPCR2_VAL
	str		r1,   [r0]

	ldr		r0,	=GPDR0
	ldr		r1,	=CFG_GPDR0_VAL
	str		r1,   [r0]

	ldr		r0,	=GPDR1
	ldr		r1,	=CFG_GPDR1_VAL
	str		r1,   [r0]

	ldr		r0,	=GPDR2
	ldr		r1,	=CFG_GPDR2_VAL
	str		r1,   [r0]

	ldr		r0,	=GAFR0_L
	ldr		r1,	=CFG_GAFR0_L_VAL
	str		r1,   [r0]

	ldr		r0,	=GAFR0_U
	ldr		r1,	=CFG_GAFR0_U_VAL
	str		r1,   [r0]

	ldr		r0,	=GAFR1_L
	ldr		r1,	=CFG_GAFR1_L_VAL
	str		r1,   [r0]

	ldr		r0,	=GAFR1_U
	ldr		r1,	=CFG_GAFR1_U_VAL
	str		r1,   [r0]

	ldr		r0,	=GAFR2_L
	ldr		r1,	=CFG_GAFR2_L_VAL
	str		r1,   [r0]

	ldr		r0,	=GAFR2_U
	ldr		r1,	=CFG_GAFR2_U_VAL
	str		r1,   [r0]

	ldr		r0,	=PSSR
	ldr		r1,	=CFG_PSSR_VAL
	str		r1,	[r0]
		
no_init_gpios:    
	
	/* The procedure below uses sample code generated by the Intel PXA250 Memory
	 * Configuration Tool [http://appzone.intel.com/pcg/pxa250/memory/] which has
	 * then been modified manually while following the documentation in
	 * "Intel PXA255 Process Developer's Manual" [order #278693-002] 
	 * Chapter 6.11 - Hardware, Watchdog, or Sleep Reset Operation */

	/* Step 1: After hardware reset, complete a power-on wait of 200us to allow clocks to stabilize */

	ldr	r3, =OSCR		/* reset the OS Timer Count to zero */
	mov	r2, #0
	str	r2, [r3]
	ldr	r4, =0x300		/* really 0x2E1 is about 200usec, so 0x300 should be plenty */
10:
	ldr	r2, [r3]
	cmp	r4, r2
	bgt	10b

/* 
 * If we're recovering a board via JTAG, don't reinitialize the memory controller,
 * as the JTAG config script has already done that.
 */
start_mem_controller_init:

    // We've determined we're not the bootloader running out of RAM at the blessed location.
    // See if we're running from RAM.  If we are, this must be a JTAG request.
    // If a JTAG request, SKIP the memory controller initialization.
    //
	cmp	pc, #CFG_DRAM_BASE      /** test if we're in SDRAM		**/
	bhs	mem_init_complete       /** if we are, then this is a JTAG recovery request, so
                                    jump wake code and go init gpio's	**/

	/* Step 1a: Write MSC0, MSC1, MSC2 */
	ldr	r3, =MSC0		/* Configures /CS0 and /CS1 */
	ldr	r2, =CFG_MSC0_VAL           /* 0x7FF8A5CB */		/* Generated value */
	str	r2, [r3]
	ldr	r2, [r3]		/* the MSC register should be read after it is written with */
					/* a new value before an access to the memory is attempted. */
					/* (see MSC section of manual) */

	ldr	r3, =MSC1		/* Configure CS2 */
	ldr	r2, =CFG_MSC1_VAL           /* 0x7FF07FF8 */
	str	r2, [r3]
	ldr	r2, [r3]

	ldr	r3, =MSC2		/* Configure CS4-5 */
	ldr	r2, =CFG_MSC2_VAL           /* 0x80088008*/
	str	r2, [r3]
	ldr	r2, [r3]

	/* Step 1b: Write MECR, MCMEM0, MCMEM1, MCATT0, MCATT1, MCIO0, MCIO1 (Skip on gumstix) */
	/* Step 1c: Write MDREFR:K0RUN and MDREFR:E0PIN. Configure MDREFR:K0DB2. Retain
	 *          the current values of MDREFR:APD and MDREFR:SLFRSH. MDREFR:DRI must
	 *          contain a valid value. Deassert MDREFR:KxFREE. (Skip on gumstix) */

#ifdef CONFIG_FIONA
	ldr	r3, =MDREFR
	ldr	r2, [r3]		    /* read MDREFR value */
	str	r2, [r3]		    /* read MDREFR value */

#if 0
//	ldr	r1, =0xffa5ffff 
	ldr	r1, =0xfffff000 
	bic	r2, r2, r1		    /* Retain APD and SLFRSH, K1DB2 and K2DB2 */
	orr	r2, r2, #0x0000000F	/* configure a valid SDRAM Refresh Interval (DRI) */
	str	r2, [r3]		    /* set valid DRI, retain APD and SLFRSH */
#endif

	ldr	r1, =0xfff
	bic	r2, r2, r1		    /* Clear DRI -- INTEL TOOL LEAVES THIS STEP OUT!! */
	orr	r2, r2, #0x00F		/* configure a valid SDRAM Refresh Interval (DRI) */
	str	r2, [r3]

	/* Step 2: Configure Synchronous Static memory (Skip on gumstix) */
    ldr r4, =SXCNFG
    mov r1, #0
    str r1, [r4]

	/* Step 4: In systems that contain SDRAM, transition the SDRAM controller through the following state sequence:
	 *         a. self-refresh
	 *         b. power-down
	 *         c. PWRDNX
	 *         d. NOP */

	orr	r2, r2, #0x00010000	/* assert K1RUN for SDCLK1 */
	str	r2, [r3]		/* change from "self-refresh and clock-stop" to "self-refresh" state */

	bic	r2, r2, #0x00400000	/* clear SLFRSH bit field */
	str	r2, [r3]		/* change from "self-refresh" to "Power-down" state */

	orr	r2, r2, #0x00008000	/* set the E1PIN bit field */
	str	r2, [r3]		/* change from "Power-down" to "PWRDNX" state */

	nop	/* no action is required to change from "PWRDNX" to "NOP" state */

#else   // USE_GUMSTIX_MEM_INIT

	ldr	r3, =MDREFR
	ldr	r2, [r3]		/* read MDREFR value */
	str	r2, [r3]		/* configure K2DB1 and K2DB2 */

	ldr	r1, =0xfff
	bic	r2, r2, r1		/* Clear DRI -- INTEL TOOL LEAVES THIS STEP OUT!! */
/* Original code used 0x18, Jateen gives spec of 0xF for Fiona */
/*	orr	r2, r2, #0x018		*//* configure a valid SDRAM Refresh Interval (DRI) */
	orr	r2, r2, #0x00F		/* configure a valid SDRAM Refresh Interval (DRI) */
	str	r2, [r3]

#endif // CONFIG_FIONA

	/* Step 4f: Write MDCNFG (with enable bits deasserted), MDCNFG:DE3:2,1:0 set to 0. */

	ldr	r3, =MDCNFG		/* Load the SDRAM Configuration register. Must not be enabled yet. */
	ldr	r2, =0x000001C8		/* Value from Intel tool */
	str	r2, [r3]		/* Write to MDCNFG register */

	/* Step 5: For systems that contain SDRAM, wait a specified NOP power-up waiting period required by
	 * the SDRAMs to ensure the SDRAMs receive a stable clock with a NOP condition */

	ldr	r3, =OSCR		/* reset the OS Timer Count to zero */
	mov	r2, #0 
	str	r2, [r3] 
	ldr	r4, =0x300		/* really 0x2E1 is about 200usec, so 0x300 should be plenty */
20:
	ldr	r2, [r3] 
	cmp	r4, r2
	bgt	20b 

	/* Step 6: Ensure the data cache is disabled -- should not be needed out of reset, but best to be safe*/

	mov	r0, #0x78		/* turn everything off */
	mcr	p15, 0, r0, c1, c0, 0	/* caches off, MMU off, etc. */

	CPWAIT				/* wait for co-processor */

	/* Step 7: On a hardware reset in systems that contain SDRAM, trigger the specified number (typically
		eight) of refresh cycles by attempting non-burst read or write accesses to any disabled
		SDRAM bank. Each such access causes a simultaneous CBR refresh cycles for all four banks,
		which causes a pass through the CBR state and back to NOP. On the first pass, the PALL state
		occurs before the CBR state. */

	ldr	r3, =PHYS_SDRAM_1
	mov	r2, #9	/* now must do 9 (!) refresh or CBR commands before the first access */
CBR_refresh1:
	str	r2, [r3]
	subs	r2, r2, #1 
	bne	CBR_refresh1 

	/* Step 8: Can re-enable DCACHE if it was disabled above (Skip on fiona) */

	/* Step 9: Enable SDRAM partitions */

	ldr	r3, =MDCNFG		/* sdram config -- sdram enable */
	ldr	r2, [r3] 
	orr	r2, r2, #0x00000001	/* enable partition 0 only */
	str	r2, [r3]		/* write to MDCNFG */

	/* Step 10: In systems containing SDRAM, write the MDMRS register to trigger an MRS command to all
	 * enabled banks of SDRAM. For each SDRAM partition pair that has one or both partitions
	 * enabled, this forces a pass through the MRS state and back to NOP. The CAS latency must be
	 * the only variable option and is derived from the value programmed in the
	 * MDCNFG:MDTC0,2 fields. The burst type is programmed to sequential and the length is set
	 * to four. */

	ldr	r3, =MDMRS		/* write the MDMRS */
/*	ldr	r2, =0x00020022*/		/* the writable bits will be written as a 0 */
	ldr	r2, =0x00220022		/* the writable bits will be written as a 0 */
	str	r2, [r3]

	/* Check the physical board_id to see if this has the
	 * mobile SDRAM that we need to drive to half-drive strength.
	 * If board_id == BOARD_REV_ID_REV3, then we configure the SDRAM
	 * EMRS register to half drive strength, full array
	 */
	ldr	r3, =PHYS_BOARD_REV_ID
	ldr	r2, [r3]
	and r2, r2, #BOARD_REV_ID_MASK  /* mask off non-significant bits */
	cmp	r2, #BOARD_REV_ID_REV3
	blt	end_of_mobile_done

	/*
	 * We have SDRAM that we need to configure to be more quiet...
	 * Set PASR to Full Array, Drive Strength to 1/2, mode EMRS for Mobiles SDRAM
	 */
	ldr	r3, =MDMRSLP		/* write the MDMRSLP */
	ldr	r2, =0x80208020
	str	r2, [r3]

end_of_mobile_done:


#if 1
	/* Step 11: (optional) Enable auto-power-down */

	ldr	r3, =MDREFR		/* enable auto-power-down */
	ldr	r2, [r3] 
	orr	r2, r2, #0x00100000	/* set the APD bit */
	str	r2, [r3]		/* write to MDREFR */
#endif

mem_init_complete:

#if defined(CONFIG_FIONA) && defined(CONFIG_POWER_SAVINGS)
    /* Did we wake from a GPIO1 event ?? */
    ldr     r0, =PEDR
    ldr     r0, [r0]
    tst     r0, #1
    beq     end_of_memsetup /* Branch if we DIDN'T just wake up */

    /* If we're waking up, branch to resume code
     * Don't resume on reset, though.
     */
    ldr     r0, =RCSR
    ldr     r1, [r0]
#ifdef CONFIG_WAKE_ON_SLEEP_ONLY
    cmp     r1, #4
    ldreq   r0, =PSPR
    ldreq   r1, [r0]
    moveq   pc, r1      /* Jump to resume pc if we just woke up */
#else
#if 0
    ldr     r2, =0x0F
    str     r2, [r0]    /* Clear all 4 causes in RCSR - we're done with it */
    ldr     r0, =PEDR   
    ldr     r2, =0x01 
    str     r2, [r0]    /* Clear GPIO0 cause in PEDR - we're done with it */
#endif
    tst     r1, #4
    ldrne   r0, =PSPR
    ldrne   r1, [r0]
    movne   pc, r1      /* Jump to resume pc if we just woke up */
#endif

#endif /* CONFIG_FIONA */

end_of_memsetup:
	mov	pc, lr 


#ifdef INCLUDE_LED_BLINKER

/* Purpose: Blink the "Heart Beat" LED
 * Inputs:  D0 - blink count
 * Trashes: r0-r4
 */
.globl blink_led
blink_led:
    mov     r3, lr
    mov     r4, r0      /* grab blink count */

    /* Make LED GPIO an output */
	ldr		r0,	=GPDR0
    ldr     r1, [r0]
    orr     r1, r1, #0x20000
    str     r1, [r0]

blink_led_loop:
    /* Turn on LED */
	ldr		r0,	=GPSR0
    ldr     r1, =0x20000
	str		r1, [r0]

    /* Wait a bit */
    bl      led_blink_wait

    /* Turn off LED */
	ldr		r0,	=GPCR0
    ldr     r1, =0x20000
	str		r1, [r0]
    
    /* Wait a bit */
    bl      led_blink_wait

    /* Loop if needed */
    subs    r4, r4, #1
    bhi     blink_led_loop

    /* Return */
    mov     pc, r3
    
led_blink_wait:
    ldr     r2, =0x03000000
led_blink_loop:
    subs    r2,r2,#1
    bhi     led_blink_loop
    mov     pc, lr

#endif // INCLUDE_LED_BLINKER


#endif	/* ifndef CONFIG_FIONA */
