/*
 * recovery.S - Recovery mode support for Flash-based systems
 *
 * Copyright (C) 2006	Nick Vaccaro  <nvaccaro@lab126.com>
 *			Mike Puckett  <mikepuck@lab126.com>
 * Copyright (C) 2005	Jerry Katzung <katzung@lab126.com>
 *
 */
#include <configs/fiona.h>

#undef  DEBUG_IOC_TRAFFIC

#ifdef CONFIG_RECOVERY_MODE

#define IOC_KEYBOARD_RECOVERY_ENABLED
#define PNLCD_OFF_DURING_FALSE_OFF

.globl	image_entry_point
image_entry_point:
	
	/* r0 contains a selector code {0,1,2,...}			*/

	adrl	r4, image_entry_point_table
	add	r4, r4, r0, LSL #2
	bx	r4
image_entry_point_table:
	b	check_recovery_mode
	b	after_check_recovery_mode
	b	check_recovery_mode

.globl	check_recovery_mode
check_recovery_mode:

	/* On entry, the recovery image is located at TEXT_BASE-BL_SIZE,
	 * and the normal boot image is located at TEXT_BASE.
	 * The bss area has not yet been initialized.
	 *
	 * This code will determine which image to run.  If the normal
	 * bootloader image is corrupt, or the user is holding down the
	 * "force recovery mode" key sequence, we will copy the recovery
	 * image to TEXT_BASE and resume execution there.  Otherwise, we
	 * will finish setting up the normal bootloader image and jump to it.
	 */
	
	/* Check for recovery mode request:
	 *	P1 and P2:	Power switch held down (GPIO 0)
	 *	P3 and later:	Magic keys from IOC
	 */

//!	mov	r4, lr			/* save return address		*/

#if defined(HAS_TRUE_POWER_SWITCH)
        /* If we're on a DVT or newer, we have a power switch that can get
         * out-of-sync.  To deal with that, we check the state and if it's OFF,
         * we put Fiona into deep sleep until they turn the switch ON.
         *
         * Make sure power switch is in correct (ON) position
         */
        bl      is_power_switch_on

#if 0   // njv FIXME - remove this
        // If power switch is OFF, skip recovery mode stuff....
        bne     crm_normal
#endif
#endif

	bl	dbg_init_serial

#ifdef CONFIG_SHOW_BOOT_PROGRESS
	// Initialize the progress bar and idicate we're at position #1
	bl	progress_bar_init
	mov	r0, #RECOVERY_PROGRESS_1
	bl	progress_bar
#endif

	ldr	sp, _RECOVER_RAM_STACK	// sp = RECOVER_RAM_STACK
	bl	init_boot_globals	// init_boot_globals()
	
	adrl	r0, dbg_str_crm_entry
	bl	dbg_puts
	
	bl	got_boot_key		// r0 = got_boot_key()
	tst	r0, #1			// if ( 0 != r0 )
	bne	crm_normal		//	goto crm_normal

	/* The recovery image is located at the base of the ROM.  If the PC
	 * is outside this range, we're either in the "normal" bootloader or
	 * we're running in RAM, so we should just return from this routine.
	 */

//!	cmp	pc, #FLASH_BOOTLOADER_BASE      /* Beyond the recovery area?*/
//!	bhi	crm_exit		        /* if so, just exit	*/

	adrl	r0, dbg_str_crm_prompt
	bl	dbg_puts

	ldr	r5, _OSCR		//  r5 = OSCR
	mov	r4, #0			//  r4 = 0
	str	r4, [r5]		// *r5 = r4
	add	r4, r4, #CFG_HZ		//  r4 = CFG_HZ
	add	r4, r4, r4		//  r4 = r4 + r4 (2 * CFG_HZ = ~2 seconds)

crm_loop:
	bl	dbg_tstc
	bne	console_data_available

#ifdef IOC_KEYBOARD_RECOVERY_ENABLED
        bl	dbg_bt_tstc
#endif
	beq	crm_1

#ifdef IOC_KEYBOARD_RECOVERY_ENABLED
	bl	dbg_bt_getc

	bl	check_boot_ioc_key	// r0 = check_boot_ioc_key(r0)
	tst	r0, #1			// if ( 0 != r0 )
	bne	crm_normal		//	goto crm_normal
	
	b	crm_1
#endif

console_data_available:
	bl	dbg_getc
	bl	dbg_putc

	bl	check_boot_key		// r0 = check_boot_key(r0)
	tst	r0, #1			// if ( 0 != r0 )
	bne	crm_normal		//	goto crm_normal

crm_1:
	ldr	r0, [r5]		// r0 = *r5
	cmp	r4, r0			// if (r4 > r0)
	bgt	crm_loop		//	goto crm_loop

crm_normal:
	bl	get_boot_recover	// r0 = get_boot_recover()
	tst	r0, #1			// if ( 0 != r0 )
	bne	crm_recover		//	goto crm_recover
	
	adrl	r0, dbg_str_crm_normal_boot
	bl	dbg_puts

	// As noted above, if we get here, we're NOT in recovery mode.  We're
	// booting normally.  But, before we just jump off to the normal
	// bootloader, we checksum it.  Specifically, we:
	//
	//     (1) Check the normal bootloader.  If it's okay, it's already
	//         in the place it needs to be, so we're done and we just
	//         jump to it.
	//
	//     (2) If the normal bootloader isn't okay, we checksum the first
	//         spare bootloader.  If it's okay, we move it to where the
	//         normal bootloader was, and we jump to it.
	//
	//     (3) If the first spare bootloader isn't okay, we checksum the
	//         the second spare bootloader.  If it's okay, we move it to
	//         where the normal bootloader was, and we jump to it.
	//
	//     (4) Finally, if the second spare bootloader isn't okay, we
	//         fall back to using the recovery bootloader.
	//
	// Note:   This strategy tacitly assumes that the recovery bootloader
	//         will always be okay.
	//
	bl	do_checksum		// r0 = do_checksum()
	tst	r0, #1			// if ( 0 != r0 )
	bne	crm_recover		//	goto crm_recover

crm_jump_to_image:

#ifdef CONFIG_SHOW_BOOT_PROGRESS
	mov	r0, #RECOVERY_PROGRESS_2
	bl	progress_bar
#endif

	// Now that the bootloader, recovery or otherwise, that we're going to
	// be running with is loaded, finish up our miniature initialization
	// process.
	//
	ldr	r0, _TEXT_BASE		// r0 = TEXT_BASE
	mov	r1, #BL_SIZE		// r1 = BL_SIZE
	bl	init_bootloader         // init_bootloader(r0, r1)
	
	/* The image to execute is located at TEXT_BASE, but the stack and
	 * malloc space still need to be allocated, and the bss section still
	 * needs to be cleared.  We jump into the image's entry point to let
	 * the image take care of the appropriate setup.
	 */
	
	mov	r0, #ENTRY_PT_AFTER_CHECK_RECOVERY
	ldr	r1, _TEXT_BASE
	add	r1, r1, #ENTRY_PT_OFFSET
	mov	pc, r1
	
//!crm_exit:
//!	adrl	r0, dbg_str_crm_exit
//!	bl	dbg_puts
//!	mov	pc, r4

	/* Everything is set up for recovery mode, so additional steps are
	 * needed to prepare for a normal boot.
	 *
	 * 1) Make a copy of this code above both images and jump to it,
	 * 2) Copy the normal image to TEXT_BASE, overwriting the recovery code,
	 * 3) Clear out the bss section for the normal image,
	 * 4) Jump into the normal boot image.
	 */
crm_recover:

#ifdef CONFIG_SHOW_BOOT_PROGRESS
	// Put up an "R" at top of PNLCD to indicate recover mode
	bl	progress_bar_recover_mode
#endif

	/* Copy the recovery image to TEXT_BASE, then continue. */
	adrl	r0, dbg_str_crm_recovering
	bl	dbg_puts
	
	ldr	r1, _TEXT_BASE		/* destination address		*/
	sub	r0, r1, #BL_SIZE	/* source start address		*/
	mov	r2, r1			/* source end address		*/
crm_recovery_loop:
	ldmia	r0!, {r3-r10}
	stmia	r1!, {r3-r10}
	cmp	r0, r2
	blt	crm_recovery_loop

	b	crm_jump_to_image

do_checksum:				// int do_checksum(void)
        mov	r4, lr			// (r0-r3 trashed)	
		
	ldr	r0, _TEXT_BASE		// r0 = TEXT_BASE
	mov	r1, #BL_SIZE		// r1 = BL_SIZE
	mov	r2, #M_BL		// r2 = M_BL (N_BL - 1)
	bl	load_bootloader		// r0 = load_bootloader(r0, r1, r2)

	mov	pc, r4			// return ( r0 )
	
	/*
	 * While it doesn't help with any of the very early bootloader code,
	 * it's still nice to be able to debug the rest of the bootloader
	 * without having to re-flash it.
	 *
	 * This routine sets up to use the recovery bootloader's RAM copy
	 * mechanism from CFG_LOAD_ADDR, and then proceeds as usual from
	 * there.
	 */
.globl	boot_from_ram
boot_from_ram:
	adrl	r0, dbg_str_crm_normal_boot
	bl	dbg_puts

	ldr	r1, _TEXT_BASE		/* destination address		*/
	ldr	r0, _CFG_LOAD_ADDR	/* source start address		*/
	mov	r2, r0			/* copy start address		*/
	add	r2, r2, #BL_SIZE	/* source end address		*/
	b	crm_recovery_loop

_RECOVER_RAM_STACK:	.word RECOVER_RAM_STACK
_CFG_LOAD_ADDR:		.word CFG_LOAD_ADDR

#ifdef CONFIG_SHOW_BOOT_PROGRESS

// PN LCD Constants
//
#define FIRST_PNLCD_SEGMENT	    0
#define LAST_PNLCD_SEGMENT	    199
#define PNLCD_ENABLE_CMD	    0x0B
#define PNLCD_SEQ_OFF_CMD	    0x41
#define PNLCD_VERT_SEQ_ON_CMD	    0x42
#define PNLCD_MISC_ON		    0x44
#define PROGRESS_BAR_BASE	    198	    // Right column, bottom segment
#define PROGRESS_BAR_TOP	    0	    // Right column, top segment
#define PROGRESS_BAR_R1		    0
#define PROGRESS_BAR_R2		    1
#define PROGRESS_BAR_R3		    3
#define PROGRESS_BAR_R4		    5
#define PROGRESS_BAR_R_SEG_COUNT    4

#define BATTERY_NOT_INSTALLED_HEADER    0xE0
#define BATTERY_NOT_INSTALLED           0x03
#define BATTERY_STATUS_CMD              0x22

#if defined(HAS_TRUE_POWER_SWITCH)
/*****************************************************************
 * Routine: is_power_switch_on
 * Purpose: 
 *  Checks power switch to assure it's in the ON direction.  
 *
 *  Sets the Z bit upon return (EQ = Switch ON, NE = Switch OFF)
 *
 * Parameters:  none
 * Returns:     none
 * Trashes:     r0
 * Assumptions: Polarity of On/Off switch is 
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 * Register Usage:
 *  r0  - Holds gpio level register from which to check power switch bit
 *****************************************************************/
is_power_switch_on:
    ldr     r0, =GPLR0
    ldr     r0, [r0]
    tst     r0, #(FIONA_DVT1_POWER_SWITCH_OFF_POLARITY<<GPIO_FIONA_DVT1_POWER_SWITCH)
    mov     pc, lr
#endif

/*****************************************************************
 * Routine: is_battery_installed
 * Purpose: 
 *  Checks with the IOC to see if a battery is installed.
 *
 * Parameters:  none
 * Returns:     r0     : Battery installed status
 *              r0 = 0 : Battery not installed
 *              r0 = 1 : Battery installed
 * Trashes:     r0, r3
 * Assumptions: IOC can be talked to through bt uart and stack is set up!
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 * Register Usage:
 *  r0  - Holds and returns battery status information
 *****************************************************************/
.globl is_battery_installed
is_battery_installed:
        stmdb   sp!, {r1-r4, lr}

#ifdef DEBUG_IOC_TRAFFIC
        adrl    r0, dbg_str_ioc_comm
	bl	dbg_puts
#endif

	mov	r0, #BATTERY_STATUS_CMD // Ask IOC for battery charge status...
	bl	dbg_bt_putc

	// Get the IOC response to the last command - make sure we get the right
	// packet type
grab_ioc_char:
        bl	dbg_bt_tstc
	beq	grab_ioc_char   // Keep looping until data is ready

	bl	dbg_bt_getc     // Go get the data
        mov     r4, r0
#ifdef DEBUG_IOC_TRAFFIC
        bl      dbg_put_hex
	bl	dbg_cr
#endif
        cmp     r4, #BATTERY_NOT_INSTALLED_HEADER
        bne     grab_ioc_char

        // We found the right packet type, grab the response
get_next_char:
        bl	dbg_bt_tstc
	beq	get_next_char   // Keep looping until data is ready

	bl	dbg_bt_getc
        mov     r4, r0

#ifdef DEBUG_IOC_TRAFFIC
        bl      dbg_put_hex
	bl	dbg_cr
        adrl    r0, dbg_str_ioc_comm
	bl	dbg_puts
#endif

        cmp     r4, #BATTERY_NOT_INSTALLED
        moveq   r0, #0          // Return 0 if NO BATTERY
        movne   r0, #1          // Return 1 if BATTERY INSTALLED

        ldmia   sp!, {r1-r4, pc}   // restore and return

/*****************************************************************
 * Routine:     progress_bar_init
 * Purpose:     Enabled the PNLCD at the IOC and then
 *              clears the PNLCD display to prepare to display 
 *              the progress bar.
 * Parameters:  none
 * Returns:     none
 * Trashes:     r0, r1, r3
 * Assumptions: This routine will not clear the PNLCD first.
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 * Register Usage:
 *  r0  - Scratch register for passing data to IOC
 *  r3  - Storage location for lr
 *****************************************************************/
.globl	progress_bar_init
progress_bar_init:

	mov	r3, lr

	// Enable the PNLCD
	mov	r0, #PNLCD_ENABLE_CMD
	bl	dbg_bt_putc

	// Clear the pnlcd progress bar
	mov	r0, #PNLCD_SEQ_OFF_CMD		// Turn off Sequential Segments
	bl	dbg_bt_putc
	mov	r0, #FIRST_PNLCD_SEGMENT
	bl	dbg_bt_putc
	mov	r0, #LAST_PNLCD_SEGMENT
	bl	dbg_bt_putc

	// Suck off the IOC response to the last command
wait_for_char1:
        bl	dbg_bt_tstc
	beq	wait_for_char1   // Keep looping until data is ready
	bl	dbg_bt_getc

	mov	pc, r3

/*********************************************************
 * Routine:     progress_bar
 * Purpose:     Displays a progress bar on the PNLCD display.
 * Parameters:  r0 - Holds progress location (1 -> 100)
 * Returns:     none
 * Trashes:     r0, r1, r2, r3
 * Assumptions: This routine will not clear the PNLCD first.
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 * Register Usage:
 *  r0  - Scratch register for passing data to IOC
 *  r2  - Storage location for progress bar position input parameter
 *  r3  - Storage location for lr
 *********************************************************/
.globl  progress_bar
progress_bar:
	mov	r3, lr

        // Move progress location since r0 gets trashed by is_power_switch_on
        mov     r1, r0

#if defined(HAS_TRUE_POWER_SWITCH) && defined(PNLCD_OFF_DURING_FALSE_OFF)
        // If power switch is OFF, just return, don't update PNLCD
        bl      is_power_switch_on
        movne   pc, r3
#endif

	subs	r1, r1, #1			// Make it 0 based
	mov	r1, r1, LSL #1			// Multiply by 2 to use single column
	mov	r0, #PROGRESS_BAR_BASE
	sub	r2, r0, r1			// Invert since base of progress bar is > top of progress bar

	// Light up the progress bar
	mov	r0, #PNLCD_VERT_SEQ_ON_CMD	// Turn on Vertically Sequential Segments
	bl	dbg_bt_putc
	mov	r0, r2				// Start at progress bar position 
	bl	dbg_bt_putc
	mov	r0, #PROGRESS_BAR_BASE		// End at progress bar base
	bl	dbg_bt_putc

	// Suck off the IOC response to the last command
wait_for_char2:
        bl	dbg_bt_tstc
	beq	wait_for_char2   // Keep looping until data is ready
	bl	dbg_bt_getc
	mov	pc, r3

/*********************************************************
 * Routine:     progress_bar_error
 * Purpose:     Displays a progress bar error indicator on the PNLCD display.
 * Parameters:  r0 - Holds progress error location (1 -> 100)
 * Returns:     none
 * Trashes:     r0, r1, r2, r3
 * Assumptions: This routine will not clear the PNLCD first.
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 * Register Usage:
 *  r0  - Scratch register for passing data to IOC
 *  r2  - Storage location for progress bar position input parameter
 *  r3  - Storage location for lr
 *********************************************************/
.globl  progress_bar_error
progress_bar_error:
	mov	r3, lr

        // Move progress location since r0 gets trashed by is_power_switch_on
        mov     r1, r0

#if defined(HAS_TRUE_POWER_SWITCH) && defined(PNLCD_OFF_DURING_FALSE_OFF)
        // If power switch is OFF, just return, don't update PNLCD
        bl      is_power_switch_on
        movne   pc, r3
#endif

	subs	r1, r1, #1			// Make it 0 based
	mov	r1, r1, LSL #1			// Multiply by 2 to use single column
	mov	r0, #PROGRESS_BAR_BASE
	sub	r2, r0, r1			// Invert since base of progress bar is > top of progress bar
	adds	r2, r2, #1			// Add one to put pixel in the error column

	// Light up the location segment in the error column of the progress bar 
	mov	r0, #PNLCD_MISC_ON		// Turn on Misc Segment
	bl	dbg_bt_putc
	mov	r0, #1				// Count is 1 segment
	bl	dbg_bt_putc
	mov	r0, r2				// Indicate error location
	bl	dbg_bt_putc

	// Suck off the IOC response to the last command
wait_for_char3:
        bl	dbg_bt_tstc
	beq	wait_for_char3   // Keep looping until data is ready
	bl	dbg_bt_getc
	mov	pc, r3

/*********************************************************
 * Routine:     progress_bar_recover_mode
 * Purpose:     Displays a "r" at the top of the PNLCD display
 *              to indicate Fiona is in recovery mode
 * Parameters:  none
 * Returns:     none
 * Trashes:     r0, r1, r8, r9
 * Assumptions: This routine will not clear the PNLCD first.
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 * Register Usage:
 *  r0  - Scratch register for passing data to IOC
 *  r8  - Storage location for lr
 *********************************************************/
// Put up an "R" at top of PNLCD to indicate recover mode
progress_bar_recover_mode:
	mov	r8, lr

#if defined(HAS_TRUE_POWER_SWITCH) && defined(PNLCD_OFF_DURING_FALSE_OFF)
        // If power switch is OFF, just return, don't update PNLCD
        bl      is_power_switch_on
        movne   pc, r8
#endif

	mov	r0, #PNLCD_MISC_ON
	bl	dbg_bt_putc
	mov	r0, #PROGRESS_BAR_R_SEG_COUNT
	bl	dbg_bt_putc
	mov	r0, #PROGRESS_BAR_R1
	bl	dbg_bt_putc
	mov	r0, #PROGRESS_BAR_R2
	bl	dbg_bt_putc
	mov	r0, #PROGRESS_BAR_R3
	bl	dbg_bt_putc
	mov	r0, #PROGRESS_BAR_R4
	bl	dbg_bt_putc

        // Suck off the response
wait_for_char5:
        bl	dbg_bt_tstc
	beq	wait_for_char5   // Keep looping until data is ready
        bl	dbg_bt_getc

	mov	pc, r8

#endif // CONFIG_SHOW_BOOT_PROGRESS

/*
 * Initialize the STUART to 115200 baud, 8n1.
 * Initialize the BTUART to   9600 baud, 8n1.
 */

#define BRG_QUOTIENT	8	/* 115200 baud divisor */

/*
 * dbg_init_serial - trashes r0,r1,r2
 */

.globl	init_btuart
.globl	dbg_init_serial
dbg_init_serial:
	ldr	r1, =0x41300004			/* CKEN |= (CKEN5_STUART | CKEN7_BTUART); */
	ldr	r0, [r1]
	orr	r0, r0, #0xa0
	str	r0, [r1]

    /* pre-load up baud rate for STUART */
    mov r2, #8              /* Setup for 115,200 */
	ldr	r1, =0x40700000			/* STIER = 0; */
	mov	r0, #0				/* STFCR = 0; */

do_init_uart:
	str	r0, [r1, #8]        		/* FCR */
	mov	r0, #0				/* IER = 0; */
	str	r0, [r1, #4]

	mov	r0, #LCR_DLAB			/* STLCR = LCR_DLAB; */
	str	r0, [r1, #12]
	mov	r0, r2				/* STDLL = BRG_QUOTIENT & 0xff; */
	str	r0, [r1]
	mov	r0, #0				/* STDLH = BRG_QUOTIENT >> 8 */
	str	r0, [r1, #4]
	mov	r0, #(LCR_WLS0|LCR_WLS1)	/* STLCR = LCR_WLS0|LCR_WLS1 */
	str	r0, [r1, #12]

	mov	r0, #IER_UUE			/* STIER = IER_UUE */
	str	r0, [r1, #4]

	/* If we've prepared the BTUART already, exit */
	cmp	r2, #0x60
	moveq	pc, lr
 
init_btuart:
        /* Enable the IOC */
        ldr     r0, iGPSR1
        ldr     r1, =0x10000
        str     r1, [r0]
        
        /* Delay ~10ms after resetting IOC to allow it time to reset */
        ldr     r1, _OSCR               /* reset the OS Timer Count to zero */
        ldr     r2, [r1]        
        add     r2, r2, #0xA000         /* really 0x2E1 is about 200usec,
                                         * so 0xA000 should be plenty     */
reset_delay:
        ldr     r0, [r1]
        cmp     r2, r0
        bgt     reset_delay 

	/* Now prepare to init BTUART */
	mov	r2, #0x60			/* Setup for 9600 baud */
	ldr	r1, =0x40200000			/* BTUART */
	mov	r0, #7				/* STFCR = 7 */
	b	do_init_uart


/*
 * dbg_putc/dbg_cr - trashes r0,r1
 * Output a single byte to the serial port.
 */

.globl dbg_cr
dbg_cr:
	mov	r0, #0x0A			/* fall through to dbg_putc */

.globl dbg_putc
dbg_putc:
	and	r0, r0, #0xFF
	ldr	r1, =0x40700014
	ldr	r1, [r1]
	tst	r1, #0x40
	beq	dbg_putc
	ldr	r1, =0x40700000
	cmp	r0, #0x0A
	str	r0, [r1]
	bxne	lr
	add	r0, r0, #3
	b	dbg_putc


/*
 * dbg_bt_putc - trashes r0,r1
 * Output a single byte to the BTUART serial port.
 */

.globl dbg_bt_putc
dbg_bt_putc:
	and	r0, r0, #0xFF
	ldr	r1, =0x40200014
	ldr	r1, [r1]
	tst	r1, #0x40
	beq	dbg_bt_putc
	ldr	r1, =0x40200000
	str	r0, [r1]
	bx	lr


/*
 * dbg_bt_tstc - trashes r0
 * Check for an available byte at the BTUART serial port.
 * Returns 1 if a byte is available, 0 otherwise.
 */

.globl dbg_bt_tstc
dbg_bt_tstc:
	ldr	r0, =0x40200014
	ldr	r0, [r0]
	ands	r0, r0, #1
	bx	lr


/*
 * dbg_bt_getc - trashes r0
 * Read a single byte from the BTUART serial port (will time out on no data)
 */

.globl dbg_bt_getc
dbg_bt_getc:
	ldr	r0, =0x40200014
	ldr	r0, [r0]
	tst	r0, #1
	beq	dbg_bt_no_data
	ldr	r0, =0x40200000
	ldr	r0, [r0]
	and	r0, r0, #0xFF
	bx	lr

dbg_bt_no_data:
    	mov 	r0, #0
    	bx  	lr


/*
 * dbg_tstc - trashes r0
 * Check for an available byte at the serial port.
 * Returns 1 if a byte is available, 0 otherwise.
 */

.globl dbg_tstc
dbg_tstc:
	ldr	r0, =0x40700014
	ldr	r0, [r0]
	ands	r0, r0, #1
	bx	lr


/*
 * dbg_getc - trashes r0
 * Read a single byte from the serial port.  (Will wait forever.)
 */

.globl dbg_getc
dbg_getc:
	ldr	r0, =0x40700014
	ldr	r0, [r0]
	tst	r0, #1
	beq	dbg_getc
	ldr	r0, =0x40700000
	ldr	r0, [r0]
	and	r0, r0, #0xFF
	bx	lr


/*
 *	dbg_puts
 *
 *	r0 - ptr to null-terminated string
 *	r1 - trashed
 *	r2 - trashed
 *	r3 - trashed
 */

.globl dbg_puts
dbg_puts:
	mov	r3, lr				/* save return address */
	mov	r2, r0				/* save string pointer */
	ldrb	r0, [r2]
dbg_puts_loop:
	cmp	r0, #0
	beq	dbg_puts_exit
	bl	dbg_putc
	ldrb	r0, [r2, #1]!
	b	dbg_puts_loop
dbg_puts_exit:
	mov	pc, r3				/* return */

/*
 *	dbg_put_hex
 *
 *	r0 - 32-bit value to print
 *	r1 - trashed
 *	r2 - trashed
 *	r3 - trashed
 */

.globl dbg_put_hex
dbg_put_hex:
	mov	r3, lr				/* save return address */
	mov	r2, r0				/* save data value */
	mov	r0, #0x30
	bl	dbg_putc
	mov	r0, #0x78
	bl	dbg_putc
	
	mov	r0, r2, LSR #28			/* get most-significant nybble */
	and	r0, r0, #0x0F
	adrl	r1, dbg_hex_digits		/* point to translation string */
	ldrb	r0, [r1, r0]
	bl	dbg_putc

	mov	r0, r2, LSR #24
	and	r0, r0, #0x0F
	adrl	r1, dbg_hex_digits
	ldrb	r0, [r1, r0]
	bl	dbg_putc

	mov	r0, r2, LSR #20
	and	r0, r0, #0x0F
	adrl	r1, dbg_hex_digits
	ldrb	r0, [r1, r0]
	bl	dbg_putc

	mov	r0, r2, LSR #16
	and	r0, r0, #0x0F
	adrl	r1, dbg_hex_digits
	ldrb	r0, [r1, r0]
	bl	dbg_putc

	mov	r0, r2, LSR #12
	and	r0, r0, #0x0F
	adrl	r1, dbg_hex_digits
	ldrb	r0, [r1, r0]
	bl	dbg_putc

	mov	r0, r2, LSR #8
	and	r0, r0, #0x0F
	adrl	r1, dbg_hex_digits
	ldrb	r0, [r1, r0]
	bl	dbg_putc

	mov	r0, r2, LSR #4
	and	r0, r0, #0x0F
	adrl	r1, dbg_hex_digits
	ldrb	r0, [r1, r0]
	bl	dbg_putc

	and	r0, r2, #0x0F
	adrl	r1, dbg_hex_digits
	ldrb	r0, [r1, r0]
	bl	dbg_putc

	mov	pc, r3

	.align	5
dbg_hex_digits:
	.byte	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37
	.byte	0x38, 0x39, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46
dbg_str_crm_entry:
	.asciz	"\ncheck_recovery: "
dbg_str_crm_prompt:
	.asciz	"shift-<r>ecover, shift-<u>pdate, shift-</> reset...\n"
//!dbg_str_crm_exit:
//!	.asciz	" exiting...\n"
dbg_str_crm_normal_boot:
	.asciz	" normal boot...\n"
dbg_str_crm_diag_boot:
	.asciz	" Running diagnostics...\n"
dbg_str_crm_recovering:
	.asciz	" recovering...\n"
#ifdef DEBUG_IOC_TRAFFIC
dbg_str_ioc_comm:
	.asciz	"IOC receive traffic...\n"
#endif

	.align	5

#endif	/* CONFIG_RECOVERY_MODE */

/* vim: set ts=8 sw=8 tw=80: */
