/* 
 * Copyright (C) 2005-2007 Lab126, Inc.
 * 
 * Nick Vaccaro,  Lab126, Inc.	<nvaccaro@lab126.com>
 * Jerry Katzung, Lab126, Inc.	<katzung@lab126.com>
 * Mike Puckett,  Lab126, Inc.	<mikepuck@amazon.com>
 *
 * Copyright (C) 2004 Gumstix, Inc.
 *
 *   Gumstix u-boot 1.1.1
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Written by Craig Hughes <craig@gumstix.com> 2004
 *
 */

#ifndef __CONFIG_H
#define __CONFIG_H

#define CONFIG_PXA250			/* this is an PXA250 CPU */
#define CONFIG_FIONA			/* config for fiona board */
#define CONFIG_POWER_SAVINGS		/* config for fiona board power management */
#undef	CONFIG_WAKE_ON_SLEEP_ONLY	/* Ignore wake event if reset also occurred */
#define CONFIG_INIT_CRITICAL		/* undef for developing */
#define CONFIG_RECOVERY_MODE		/* recovery is possible */
#undef	CONFIG_USE_IRQ			/* don't need use IRQ/FIQ */
#define CONFIG_SHOW_BOOT_PROGRESS	/* use the PNLCD as a boot progress bar */
#define CONFIG_ENV_OVERWRITE		/* allow things like serial# env variable to be overwritten */
#define CONFIG_EINK_APOLLO		/* say that we have an Apollo-based Eink controller */
#define CONFIG_CHECK_BOOTBLOCK		/* check to see whether bootblock is stale or not */
#define CONFIG_CHECK_BATTERY_LEVEL	/* check to see whether we have enough battery power to boot */
#define CONFIG_USE_ALERT_PICTURES	/* use pictures for on-screen alerts */
#undef  CONFIG_CALCULATE_OP_AMP_OFFSET  /* don't calculate on the fly, it's not reliable */
#define CONFIG_USE_STATIC_WATERMARKS	/* use static watermarks instead of from the environment */

#ifndef __LITTLE_ENDIAN
#define __LITTLE_ENDIAN 1234		/* For some reason this is not defined */
#endif

/*
 * Select serial console configuration: STUART is default on Fiona
 */
#define CONFIG_STUART
#define CONFIG_BAUDRATE		115200
#define CONFIG_SILENT_CONSOLE

/*
 * Definition of u-boot built-in commands. Check out CONFIG_CMD_DFL if
 * neccessary in include/cmd_confdefs.h file. (Un)comment for getting
 * functionality or size of u-boot code.
 */
#define CONFIG_COMMANDS	(  CONFIG_CMD_DFL	\
			& ~CFG_CMD_IMLS		\
			& ~CFG_CMD_NET 		\
			& ~CFG_CMD_NFS		\
			& ~CFG_CMD_SETGETDCR	\
			& ~CFG_CMD_AUTOSCRIPT	\
			| CFG_CMD_BML		\
			| CFG_CMD_CACHE 	\
			| CFG_CMD_FAT		\
			| CFG_CMD_FIONA		\
			| CFG_CMD_MMC		\
			| CFG_CMD_NAND		)

#include <cmd_confdefs.h>

/*
 * Boot options. Setting delay to -1 stops autostart count down.
 * NOTE: Sending parameters to kernel depends on kernel version and
 * 2.4.19-rmk6-pxa1 patch used while my u-boot coding didn't accept 
 * parameters at all! Do not get confused by them so.
 */
#define CONFIG_IDENT_STRING	"\n\n*** Welcome to Kindle ***"
#define CONFIG_MISC_INIT_R	/* misc_init_r() sets Fiona board serial no. */

/*
 * User Diagnostic Constants
 */
#define USER_DIAGNOSTIC_CODE_SIGNATURE		0x6b63694E
#define USER_DIAGNOSTIC_SIGNATURE_OFFSET	0x0
#define USER_DIAGNOSTIC_VECTOR_TABLE_OFFSET     0x40
#define USER_DIAGNOSTIC_SIGNATURE_SIZE		0x20

/*
 * To load code so that signature header is not in actual code space,
 * uncomment the code below.  Based on how Foxconn did their diag code,
 * we're not doing this as they included the signature in their codespace.
 */
//#define USER_DIAGNOSTIC_RAM_BASE		(0xA2000000-USER_DIAGNOSTIC_SIGNATURE_SIZE) 
#define USER_DIAGNOSTIC_RAM_BASE		0xA2000000 
#define USER_DIAGNOSTIC_CODE_ENTRY_PT		(USER_DIAGNOSTIC_RAM_BASE+USER_DIAGNOSTIC_SIGNATURE_OFFSET+USER_DIAGNOSTIC_SIGNATURE_SIZE)
#define USER_DIAGNOSTIC_VECTOR_TABLE_ADDR	(USER_DIAGNOSTIC_RAM_BASE+USER_DIAGNOSTIC_VECTOR_TABLE_OFFSET)

#define	CONFIG_RECOVER_CONSOLE_KEY		'R'	// Force recovery bootloader into console.

#define CONFIG_DIAGS_KEY			'D'	// Go into diagnostics mode (if CONFIG_BOOT_DIAGS defined).
#define CONFIG_DIAGS_IOC_KEY			0x23	// Key code for 'd'.

#define CONFIG_FW_RESET_KEY			'?'	// Force recovery bootloader to do a firmware reset.
#define CONFIG_FW_RESET_IOC_KEY			0x27	// Key code for '/'

#define CONFIG_SW_UPDATE_KEY			'U'	// Force recovery bootloader to do a firmware update.
#define CONFIG_SW_UPDATE_IOC_KEY		0x22	// Key code for 'u'.

#define CONFIG_FACTORY_UPDATE_KEY		'%'	// Force recovery bootloader to do a factory update.
#define CONFIG_FACTORY_UPDATE_IOC_KEY		0x20	// Key code for '5'.

#define CONFIG_MENU_SCROLL_IOC_KEY		0x07	// Force recovery bootloader into service menu.
#define CONFIG_MENU_HOME_IOC_KEY		0x2F	// Force recovery bootloader into service menu.
#define CONFIG_MENU_KEY				0x1B	// Force recovery bootloader into service menu (ESC, ^[).

#define CONFIG_MENU_DIAG_BOOT_KEY		'0'	// Service menu's diag selection key.
#define CONFIG_MENU_DIAG_BOOT_IOC_KEY		0x19	// Key code for '0'.

#define CONFIG_MENU_UPDATE_BOOT_KEY		'1'	// Service menu's firware update selection key.
#define CONFIG_MENU_UPDATE_BOOT_IOC_KEY		0x00	// Key code for '1'

#define CONFIG_MENU_RESET_BOOT_KEY		'2'	// Service menu's firmware reset selection key.
#define CONFIG_MENU_RESET_BOOT_IOC_KEY		0x08	// Key code for '2'

#define CONFIG_MENU_STANDARD_BOOT_KEY		'3'	// Service menu's normal boot (exit) selection key.
#define CONFIG_MENU_STANDARD_BOOT_IOC_KEY	0x10	// Key code for '3'

#define CONFIG_BOOTARGS		"console=ttyS2,115200n8 root=/dev/rd/0 initrd=/linuxrc rw reboot=hard"
#define CONFIG_BOOTCOMMAND	"bootf default"

#define CONFIG_BOOTDELAY	0		/* in seconds */
#define CFG_HUSH_PARSER
#define CFG_AUTO_COMPLETE
#define CFG_PROMPT_HUSH_PS2	"> "

/*
 * General options for u-boot. Modify to save memory foot print
 */
#define CFG_LONGHELP				/* undef saves memory	*/
#define CFG_PROMPT		"KINDLE> "	/* prompt string	*/
#define CFG_CBSIZE		512		/* console I/O buffer	*/
#define CFG_PBSIZE (CFG_CBSIZE+sizeof(CFG_PROMPT)+16)	/* print buffer size */
#define CFG_MAXARGS		16		/* max command args	*/
#define CFG_BARGSIZE		CFG_CBSIZE	/* boot args buf size	*/

#define CONFIG_CRC32_VERIFY
#define CONFIG_LOOPW
#define CONFIG_MX_CYCLIC

//
// Don't calculate on the fly, it's not reliable
//
#undef CONFIG_CALCULATE_OP_AMP_OFFSET  /* calculate op amp offset value */

/*
 * Definitions related to passing arguments to kernel.
 */
#define CONFIG_CMDLINE_TAG		/* send commandline to kernel	*/
#undef  CONFIG_SETUP_MEMORY_TAGS	/* do not send memory definition to kernel */
#define CONFIG_INITRD_TAG		/* send initrd params		*/
#define CONFIG_SERIAL_TAG		/* send serialnr from flash 	*/
#define CONFIG_REVISION_TAG		/* send board revision number from board ID register */
#define CONFIG_BOARD_RESISTANCE_TAG	/* send board resistance value */
#undef  CONFIG_VFD			/* do not send framebuffer setup */

#define CFG_MEMTEST_START	0xa1000000	/* memtest test area	*/
#define CFG_MEMTEST_END		0xa2000000

#undef  CFG_CLKS_IN_HZ			/* use HZ for freq. display */
#define CFG_HZ			3686400	/* increments at 3.6864 MHz */
#define CFG_BAUDRATE_TABLE	{ 9600, 19200, 38400, 57600, 115200, 230400 }
#define CFG_CONSOLE_INFO_QUIET

/*
 * GPIO Macros
 */
#define GPIO_SET1BIT(gpionum, val) ((val & 0x1)  << ((gpionum)%32))
#define AFSET(gpionum, val) ((val & 0x3) << (((gpionum)%16)*2))
#define BITSET(gpionum) (GPIO_SET1BIT((gpionum), 1))

/*
 * Build-specific constants
 */
#if defined(BUILD_FOR_DVT1_HARDWARE) || defined(BUILD_FOR_REV2_HARDWARE)
#include <configs/fiona_dvt1.h>
#else
#error "Kindle Board Revision is not supported.  Make sure BUILD_FOR_... is set properly."
#endif


/*************************************************************************
 * GPIO Constants
 ************************************************************************/

#define GPIO_FIONA_IOC_RESET        48
#define GPIO_FIONA_IOC_RESET_MD     ( GPIO_FIONA_IOC_RESET | GPIO_OUT )

#define GPIO_FIONA_TEMP_SENSOR      51
#define GPIO_FIONA_TEMP_SENSOR_MD   ( GPIO_FIONA_TEMP_SENSOR | GPIO_OUT | GPIO_DFLT_LOW )

#define GPIO57_MMC_POWER            57
#define GPIO57_MMC_POWER_MD         ( GPIO57_MMC_POWER | GPIO_OUT )

#define GPIO_FIONA_APOLLO_COMMON    80
#define GPIO_FIONA_APOLLO_COMMON_MD ( GPIO_FIONA_APOLLO_COMMON | GPIO_OUT | GPIO_DFLT_LOW )

#define GPIO_FIONA_APOLLO_POWER     82
#define GPIO_FIONA_APOLLO_POWER_MD  ( GPIO_FIONA_APOLLO_POWER | GPIO_OUT | GPIO_DFLT_HIGH )

#define GPIO24_FIONA_P3_WAN_ON_OFF  24
#define GPIO81_FIONA_P3_MODULE_WAKE 81

/*
 * SDRAM Memory Map
 */
#define CONFIG_NR_DRAM_BANKS	1		/* we have 1 bank of SDRAM */
#define PHYS_SDRAM_1		0xa0000000	/* SDRAM bank #1	   */
#define PHYS_SDRAM_1_SIZE	0x04000000	/* 64 MB ( 2 chip )	   */

#define CFG_DRAM_BASE		PHYS_SDRAM_1
#define CFG_DRAM_SIZE		PHYS_SDRAM_1_SIZE

/*
 * Configuration for FLASH memory
 */

#define CONFIG_FLASH_SAMSUNG_ONENAND		/* Include OneNAND support */
#define CONFIG_ONENAND_SYNC	false		/* true to default on, false to default off */

#define CONFIG_FLASH_RETRIES	4
#define CONFIG_FLASH_PAUSE	5

#define CONFIG_ENV_VERS		20070807	/* Bump when environment changes */
#define CONFIG_EXTRA_ENV_SETTINGS				\
	"env_vers="		MK_STR(CONFIG_ENV_VERS)		"\0"

#define ENV_NAME_SIZE		32
#define ENV_NUM_SIZE		9

#define ENV_VERS		"env_vers"
#define ENV_VERS_STR_SZ		ENV_NUM_SIZE
#define ENV_VERS_BAD		true
#define ENV_VERS_GOOD		false

#define ONENAND_SYNC_RD_ENV	"onenand_sync_reads"
#define ONENAND_SYNC_RD_STR_SZ	6

#define HW_FLAGS_ENV		"hw_flags"
#define HW_FLAGS_ENV_STR_SZ	ENV_NUM_SIZE

#define BAT_LEVEL_0_ENV		"bat_level#0"
#define BAT_LEVEL_1_ENV		"bat_level#1"
#define BAT_LEVEL_2_ENV		"bat_level#2"
#define BAT_LEVEL_3_ENV		"bat_level#3"
#define BAT_LEVEL_ENV_STR_SZ	5

#define DEV_ENV                 "dev"
#define DEV_ENV_TRUE		"true"
#define DEV_ENV_FPOW		"fpow"
#define DEV_ENV_EINK		"eink"
#define DEV_ENV_STR_SZ		6

#define ENV_SCRIPT		"env_script"
#define ENV_SCRIPT_STR_SZ	256

#define ENV_OP_AMP		"op_amp#"
#define ENV_OP_AMP_STR_SZ	ENV_NUM_SIZE

#define ENV_GAIN		"gain#"
#define ENV_GAIN_STR_SZ		ENV_NUM_SIZE

#define ENV_RVE			"rve#"
#define ENV_RVE_STR_SZ		ENV_NUM_SIZE

#define PHYS_FLASH_1		0x00000000	/* Flash Bank #1  */
#define PHYS_FLASH_2		0x04000000	/* Flash Bank #2  */
#define PHYS_FLASH_3		0x08000000	/* Flash Bank #3  */
#define PHYS_FLASH_SECT_SIZE	0x00020000	/* 128 KB sectors */
#define PHYS_FLASH_PAGE_SIZE	0x00000800	/*   2 KB pages   */

#define CFG_FLASH_BANKS_LIST	{ PHYS_FLASH_1, PHYS_FLASH_2, PHYS_FLASH_3 }
#define CFG_FLASH_BASE		PHYS_FLASH_1
#define CFG_FLASH_CFI				/* flash is CFI conformant    */
//#define CFG_FLASH_CFI_DRIVER			/* use common cfi driver      */
#define CFG_FLASH_USE_BUFFER_WRITE 1		/* buffered writes 20x faster */
#define CFG_MAX_FLASH_BANKS	3		/* max # of memory banks      */
#define CFG_MAX_FLASH_SECT	1024		/* max sectors on one chip    */
#define CFG_FLASH_PROTECTION			/* use hw flash protection    */

#define CFG_FLASH_ERASE_TOUT	(25*CFG_HZ)
#define CFG_FLASH_WRITE_TOUT	(25*CFG_HZ)

/*
 * Flash "partitioning" layout
 */

#define FL_BASE			(CFG_FLASH_BASE)		/* 0x00000000 */
#define FL_BLK			(PHYS_FLASH_SECT_SIZE)		/*      128KB */
#define FL_1MB			(FL_BLK * 8)			/*        1MB */
#define FL_1_BANK		(CFG_MAX_FLASH_SECT * FL_BLK)	/*	128MB */
#define FL_2_BANK		(FL_1_BANK * 2)			/*	256MB */

// 0x00000000, 256K
//
#define BL_SIZE			(2 * FL_BLK)
#define N_BL			(4)
#define M_BL			(N_BL - 1)

#define FLASH_BOOTLOADER1_BASE	(FL_BASE)
#define FLASH_BOOTLOADER1_SIZE	(BL_SIZE)

// 0x00040000, 256K
//
#define FLASH_BOOTLOADER2_BASE	(FLASH_BOOTLOADER1_BASE + FLASH_BOOTLOADER1_SIZE)
#define FLASH_BOOTLOADER2_SIZE	(BL_SIZE)

// 0x00080000, 256K
//
#define FLASH_BOOTLOADER3_BASE	(FLASH_BOOTLOADER2_BASE + FLASH_BOOTLOADER2_SIZE)
#define FLASH_BOOTLOADER3_SIZE	(BL_SIZE)

// 0x000C0000, 256K
//
#define FLASH_BOOTLOADER4_BASE	(FLASH_BOOTLOADER3_BASE + FLASH_BOOTLOADER3_SIZE)
#define FLASH_BOOTLOADER4_SIZE	(BL_SIZE)

// 0x00100000, 512K
//
#define FLASH_DIAGNOSTICS_BASE	(FLASH_BOOTLOADER4_BASE + FLASH_BOOTLOADER4_SIZE)
#define FLASH_DIAGNOSTICS_SIZE  (4 * FL_BLK)

// 0x00180000, 1536K
//
#define FLASH_KERNEL1_BASE	(FLASH_DIAGNOSTICS_BASE + FLASH_DIAGNOSTICS_SIZE)
#define FLASH_KERNEL1_SIZE	(12 * FL_BLK)

// 0x00300000, 1536K
//
#define FLASH_KERNEL2_BASE	(FLASH_KERNEL1_BASE + FLASH_KERNEL1_SIZE)
#define FLASH_KERNEL2_SIZE	(12 * FL_BLK)

// 0x00480000, 1280K
//
#define FLASH_INITRD1_BASE	(FLASH_KERNEL2_BASE + FLASH_KERNEL2_SIZE)
#define FLASH_INITRD1_SIZE 	(10 * FL_BLK)

// 0x005C0000, 1280K
//
#define FLASH_INITRD2_BASE	(FLASH_INITRD1_BASE + FLASH_INITRD1_SIZE)
#define FLASH_INITRD2_SIZE 	(10 * FL_BLK)

// 0x00700000, 14MB
//
#define FLASH_FILESYSTEM_BASE	(FLASH_INITRD2_BASE + FLASH_INITRD2_SIZE)
#define FLASH_FILESYSTEM_SIZE	(14 * FL_1MB)

// 0x01500000, 15MB
//
#define FLASH_FILESYSTEM1_BASE  (FLASH_FILESYSTEM_BASE + FLASH_FILESYSTEM_SIZE)
#define FLASH_FILESYSTEM1_SIZE  (15 * FL_1MB)

// 0x02400000, 22MB
//
#define FLASH_FILESYSTEM2_BASE  (FLASH_FILESYSTEM1_BASE + FLASH_FILESYSTEM1_SIZE)
#define FLASH_FILESYSTEM2_SIZE  (22 * FL_1MB)

// 0x03A00000, 192M
//
#define FLASH_FILESYSTEM3_BASE	(FLASH_FILESYSTEM2_BASE + FLASH_FILESYSTEM2_SIZE)
#define FLASH_FILESYSTEM3_SIZE	(192 * FL_1MB)

// 0x0FA00000, 256K
//
#define FLASH_ENVIRONMENT_BASE	(FLASH_FILESYSTEM3_BASE + FLASH_FILESYSTEM3_SIZE)
#define FLASH_ENVIRONMENT_SIZE	(2 * FL_BLK)

// 0x0FA40000, 5888K
//
#define FLASH_RESERVED_BASE	(FLASH_ENVIRONMENT_BASE + FLASH_ENVIRONMENT_SIZE)
#define FLASH_RESERVED_SIZE	(FL_2_BANK - FLASH_RESERVED_BASE)

// 0x00700000, 249M
//
#define FLASH_ROOTIMAGE_BASE	(FLASH_FILESYSTEM_BASE)
#define FLASH_ROOTIMAGE_SIZE	(FL_2_BANK - FLASH_FILESYSTEM_BASE)

#define FLASH_DIAGNOSTICS_END   (FLASH_DIAGNOSTICS_BASE+FLASH_DIAGNOSTICS_SIZE)

// EVT1's and newer use JFFS2_NAND
#define CONFIG_JFFS2_NAND

/*
 * This information is needed for NAND flash support in U-Boot.
 * These defines are necessary if CFG_CMD_NAND is enabled.
 */

#define CFG_MAX_NAND_DEVICE	2
#define CFG_NAND_BASE		PHYS_FLASH_1
#define SECTORSIZE		512
#define ADDR_COLUMN		1
#define ADDR_PAGE		2
#define ADDR_COLUMN_PAGE	3
#define NAND_ChipID_UNKNOWN	0x00
#define NAND_MAX_FLOORS		1
#define NAND_MAX_CHIPS		2

#define NAND_WAIT_READY(nand)
#define WRITE_NAND_COMMAND(d, adr)
#define WRITE_NAND_ADDRESS(d, adr)
#define WRITE_NAND(d, adr)
#define READ_NAND(adr)		0
#define NAND_DISABLE_CE(nand)
#define NAND_ENABLE_CE(nand)
#define NAND_CTL_CLRALE(nandptr)
#define NAND_CTL_SETALE(nandptr)
#define NAND_CTL_CLRCLE(nandptr)
#define NAND_CTL_SETCLE(nandptr)

/*
 * Blessed location to load Bootloader to RAM.  This address
 * is used to determine if the running bootlaoder is the blessed bootloader
 * running from RAM, a JTAG bootloader running from RAM, or a bootloader
 * running from ROM.
 */
#define BLESSED_RAM_LOCATION    (CFG_DRAM_BASE+CFG_DRAM_SIZE - 0x00100000)		// 0xA3F00000

/*
 * To communicate between the bootloader and kernel between boot-ups, we reserve
 * 4K (PAGE_SIZE) of RAM.
 */
#define RESERVE_RAM_SIZE	4096							// 0x00001000
#define BOOT_GLOBALS_START	(CFG_DRAM_BASE+CFG_DRAM_SIZE - RESERVE_RAM_SIZE)	// 0xA3FFF000

#define FRAMEBUFFER_SIZE	(4096 * 30)						// 0x0001E000
#define FRAMEBUFFER_START	(BOOT_GLOBALS_START - FRAMEBUFFER_SIZE)			// 0xA3FE1000

/*
 * As noted above, when we're first booting up, we're checking the state of the
 * bootloader.  In order to call some C routines from within the early bootloader's
 * assembly-oriented world, we need to set up a stack.  We set the stack 256 bytes
 * below the boot globals to prevent a stack overflow from wiping out the boot
 * globals.
 *
 * Note:  The recovery bootloader loads all 4 bootloaders from 0xA3EC0000 to 0xA3FC0000.
 */
#define RECOVER_RAM_STACK	(CFG_DRAM_BASE+CFG_DRAM_SIZE - RESERVE_RAM_SIZE - 256)	// 0xA3FFEFF0

/*
 * Environment setup. Definitions of monitor location and size with
 * definition of environment setup ends up in 2 possibilities.
 * 1. Embedded environment - in u-boot code is space for environment
 * 2. Environment is read from predefined sector of flash
 * Right now we support 2. possiblity, but expecting no env placed
 * on mentioned address right now. This also needs to provide whole
 * sector for it - for us 256Kb is really waste of memory. U-boot uses
 * default env. and until kernel parameters could be sent to kernel
 * env. has no sense to us.
 */
#define CFG_ENV_IS_IN_BBML
#define CFG_ENV_SIZE		5120
#define CFG_ENV_BUFFER_BASE	(CFG_MEMTEST_START - FLASH_ENVIRONMENT_SIZE)

/*
 * Malloc pool need to host env + 128 Kb reserve for other allocations.
 */
#define CFG_MALLOC_LEN		(CFG_ENV_SIZE + (128<<10) )
#define CONFIG_STACKSIZE	(120<<10)	/* stack size */
#define CFG_GBL_DATA_SIZE	512		/* num bytes initial data       */

/*
 * Among other things, CFG_LOAD_ADDR is used as the address to boot the kernel
 * and CFG_LOAD_ADDR_ALT is used for the kernel's initrd.  We put CFG_LOAD_ADDR
 * at CFG_MEMTEST_END and CFG_LOAD_ADDR_ALT above CFG_LOAD_ADDR by the size
 * of the filesystem plus 1 block (128K).
 */
#define CFG_LOAD_ADDR		CFG_MEMTEST_END
#define CFG_LOAD_ADDR_ALT	(CFG_LOAD_ADDR + FLASH_FILESYSTEM_SIZE + FL_BLK)

/*
 * Undef CONFIG_USE_DEFAULT_BOARD_RESISTANCE to calculate board resistance
 * on-the-fly.
 *
 * define CONFIG_USE_DEFAULT_BOARD_RESISTANCE to always use board's default
 * resistance value (don't attempt calculation of board resistance)
 */
#define CONFIG_USE_DEFAULT_BOARD_RESISTANCE

/*
 * These settings are for JFFS2 support in u-boot.
 * NOTE: Enable CFG_CMD_JFFS2 for JFFS2 support.
 */

//#define CFG_JFFS_CUSTOM_PART		1

//#define CFG_JFFS2_FIRST_BANK		2
//#define CFG_JFFS2_FIRST_SECTOR	4
//#define CFG_JFFS2_NUM_BANKS		1

/* PCFR initialization */
#define CFG_PCFR_VAL    PCFR_OPDE       // Allow 3 MHz clock to sleep
#define CFG_OSCC_VAL    OSCC_OON        // Fire up 32Khz clock

/* MMC information */
#define CONFIG_MMC		1
#define CFG_MMC_BASE		0xF0000000
#define CONFIG_DOS_PARTITION

/*
 * RAW Board revision ID register at 14000000
 */
#define PHYS_BOARD_REV_ID	0x14000000	/* Board revision ID */
#define BOARD_REV_ID_MASK	0x000F0000	/* ID contained on D[19:16] */
#define BOARD_REV_ID_SHIFT	16		/* Shift right 16 bits for ID */
#define BOARD_REV_ID_P1		0x00000000
#define BOARD_REV_ID_P2		0x00000000
#define BOARD_REV_ID_P3		0x00010000
#define BOARD_REV_ID_P3_5	0x00020000
#define BOARD_REV_ID_EVT1	0x00030000
#define BOARD_REV_ID_EVT2	0x00040000
#define BOARD_REV_ID_EVT3	0x00050000
#define BOARD_REV_ID_DVT1	0x00060000
#define BOARD_REV_ID_REV2	0x00070000
#define BOARD_REV_ID_REV3	0x00080000	/* REV2 w/ KAM561633G SDRAM */

/*
 * Translated Version Constants
 */

#define FIONA_BOARD_REV3	10
#define FIONA_BOARD_REV2	9
#define FIONA_BOARD_DVT1	8
#define FIONA_BOARD_EVT3	7
#define FIONA_BOARD_EVT2	6
#define FIONA_BOARD_EVT1	5
#define FIONA_BOARD_P3_5	4
#define FIONA_BOARD_P3		3
#define FIONA_BOARD_P2		2

#define MAX_SUPPORTED_BOARD_REVS	9

/*
 * Constants that would be true/false, but including <common.h> from here is difficult.
 */
#define FIONA_VERBOSE_REVISION	1	/* Use printf()s in get_fiona_board_revision_str().	*/
#define FIONA_QUIET_REVISION	0	/* Don't use printf()s. */

#define NO_SLEEP_PSPR   	0xDEADBEEF

/*
 * Constants for progress_bar to keep the .S and .c files in sync.
 */
#define RECOVERY_PROGRESS_1	1 // Used in recovery.S
#define RECOVERY_PROGRESS_2	2 //

#define BOOTBLOCK_PROGRESS_1	1 // Used in bootblock.S
#define BOOTBLOCK_PROGRESS_2	2 //
#define BOOTBLOCK_PROGRESS_3	3 //
#define BOOTBLOCK_PROGRESS_4	4 //

#define PROGRESS_BAR_END	BOOTBLOCK_PROGRESS_4

/*
 * Constants for default board resistance values
 */
#define MIN_BOARD_RESISTANCE_DVT1       150
#define MAX_BOARD_RESISTANCE_GAIN_DVT1  60

#define MAX_BOARD_RESISTANCE_DVT1       (MIN_BOARD_RESISTANCE_DVT1 + MAX_BOARD_RESISTANCE_GAIN_DVT1)
#define DEFAULT_BOARD_RESISTANCE_DVT1   MIN_BOARD_RESISTANCE_DVT1

#define MIN_BOARD_RESISTANCE_REV2       150
#define MAX_BOARD_RESISTANCE_GAIN_REV2  60

#define MAX_BOARD_RESISTANCE_REV2       (MIN_BOARD_RESISTANCE_REV2 + MAX_BOARD_RESISTANCE_GAIN_REV2)
#define DEFAULT_BOARD_RESISTANCE_REV2   MIN_BOARD_RESISTANCE_REV2

// C-usage, starts in board.c, finishes in cmd_bootm.c.
//
#define UPDATE_PROGRESS_START	PROGRESS_BAR_END
#define MIN_PROGRESS		1
#define MAX_PROGRESS		100

#define SHOW_BOOT_PROGRESS_MAX	16

#endif  /* __CONFIG_H */

/* vim: set ts=8 sw=8 tw=80: */
