/*
 * (C) Copyright 2005-2006, Lab126, Inc.
 * Jerry Katzung, katzung@lab126.com
 * Mike Puckett, mikepuck@amazon.com
 *
 * (C) Copyright 2002
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * (C) Copyright 2002
 * Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Marius Groeger <mgroeger@sysgo.de>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <command.h>
#include <malloc.h>
#include <devices.h>
#include <version.h>
#include <net.h>

#ifdef CONFIG_DRIVER_SMC91111
#include "../drivers/smc91111.h"
#endif
#ifdef CONFIG_DRIVER_LAN91C96
#include "../drivers/lan91c96.h"
#endif

#if (CONFIG_COMMANDS & CFG_CMD_NAND)
void nand_init (void);
#endif

#if (CONFIG_COMMANDS & CFG_CMD_BML)
extern void BBM_INIT_DEVICE(void);
#endif

#ifdef CONFIG_EINK_APOLLO
extern void eink_fb_probe(void);
#endif

ulong monitor_flash_len;

#ifdef CONFIG_HAS_DATAFLASH
extern int  AT91F_DataflashInit(void);
extern void dataflash_print_info(void);
#endif

#ifndef CONFIG_IDENT_STRING
#define CONFIG_IDENT_STRING ""
#endif

const char version_string[] =
	U_BOOT_VERSION" (" __DATE__ " - " __TIME__ ")"CONFIG_IDENT_STRING;

#ifdef CONFIG_DRIVER_CS8900
extern void cs8900_get_enetaddr (uchar * addr);
#endif

#ifdef CONFIG_DRIVER_RTL8019
extern void rtl8019_get_enetaddr (uchar * addr);
#endif

#ifdef CONFIG_SHOW_BOOT_PROGRESS 
extern void update_progress_bar(void); 
#else
#define update_progress_bar() do {} while(0)
#endif

/*
 * Begin and End of memory area for malloc(), and current "brk"
 */
static ulong mem_malloc_start = 0;
static ulong mem_malloc_end = 0;
static ulong mem_malloc_brk = 0;

static
void mem_malloc_init (ulong dest_addr)
{
	mem_malloc_start = dest_addr;
	mem_malloc_end = dest_addr + CFG_MALLOC_LEN;
	mem_malloc_brk = mem_malloc_start;

	memset ((void *) mem_malloc_start, 0,
			mem_malloc_end - mem_malloc_start);
}

void *sbrk (ptrdiff_t increment)
{
	ulong old = mem_malloc_brk;
	ulong new = old + increment;

	if ((new < mem_malloc_start) || (new > mem_malloc_end)) {
		return (NULL);
	}
	mem_malloc_brk = new;

	return ((void *) old);
}

/************************************************************************
 * Init Utilities                                                       *
 ************************************************************************
 * Some of this code should be moved into the core functions,
 * or dropped completely,
 * but let's get it working (again) first...
 */

static int init_baudrate (void)
{
	DECLARE_GLOBAL_DATA_PTR;

	uchar tmp[64];	/* long enough for environment variables */
	int i = getenv_r ("baudrate", tmp, sizeof (tmp));
	gd->bd->bi_baudrate = gd->baudrate = (i > 0)
			? (int) simple_strtoul (tmp, NULL, 10)
			: CONFIG_BAUDRATE;

	return (0);
}

#ifdef CONFIG_FIONA
#define FIONA_INFO_BANNER 0
#define BOOTLOADER_BANNER 1

extern unsigned int get_fiona_board_revision_str(char *str, bool verbose);
extern void check_bootloader_status(void);
extern unsigned long get_ioc_version(void);
extern unsigned long ioc_version;

static void display_fiona_info(int which_str, bool verbose)
{
        char str[32];
        char *bootv = NULL;
        int  rev = get_fiona_board_revision_str(str, verbose);
        int  mhz = (CFG_CPUSPEED == 0x161) ? 400 : 200;
        unsigned long version = 0;

#ifdef BUILD_FOR_REV2_HARDWARE
        bootv = "REV3";
#elif BUILD_FOR_DVT1_HARDWARE
        bootv = "DVT1";
#elif BUILD_FOR_EVT3_HARDWARE
        bootv = "EVT3";
#elif BUILD_FOR_EVT2_HARDWARE
        bootv = "EVT2";
#elif BUILD_FOR_EVT1_HARDWARE
        bootv = "EVT1";
#elif BUILD_FOR_P35_HARDWARE
        bootv = "P3.5";
#else
#error "NO BUILD_FOR_XXX_HARDWARE defined!!!"
#endif
        switch ( which_str )
        {
        	case FIONA_INFO_BANNER:
        		printf("Kindle %d MHz %s Board, Revision #%x\n",
        			mhz, str, rev);
        	break;
        	
        	case BOOTLOADER_BANNER:
        		printf("\n%s Bootloader: Kindle %d MHz %s Board, Revision #%x\n",
        			bootv, mhz, str, rev);
        		
        		check_bootloader_status();

                // Get version of IOC firmware
                version = get_ioc_version();
                printf(" IOC Firmware Version #%d\n",version);

        		putc('\n');
        	break;
        }
}

static int do_fiona_info(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	int result = FAILURE;
	
	if ( 1 == argc )
	{
		display_fiona_info(FIONA_INFO_BANNER, FIONA_QUIET_REVISION);
		result = SUCCESS;
	}
	else
		printf("Usage:\n%s\n", cmdtp->usage);
	
	return ( result );	
}
#endif

static int display_banner (void)
{
	printf ("\n\n%s\n\n", version_string);
	printf ("U-Boot code: %08lX -> %08lX  BSS: -> %08lX\n",
		_armboot_start, _bss_start, _bss_end);

#ifdef CONFIG_FIONA
	display_fiona_info(BOOTLOADER_BANNER, FIONA_VERBOSE_REVISION);
#endif

	return (0);
}

static void display_config(int banknum, ulong address, ulong size)
{
	printf(" Bank #%d: 0x%08lX ", banknum, address);
	print_size(size, "\n");
}

/*
 * WARNING: this code looks "cleaner" than the PowerPC version, but
 * has the disadvantage that you either get nothing, or everything.
 * On PowerPC, you might see "DRAM: " before the system hangs - which
 * gives a simple yet clear indication which part of the
 * initialization is failing.
 */
static int display_dram_config (void)
{
	DECLARE_GLOBAL_DATA_PTR;
	int i;

	puts("RAM Configuration:\n");

	for( i = 0; i < CONFIG_NR_DRAM_BANKS; i++)
		display_config(i, gd->bd->bi_dram[i].start, gd->bd->bi_dram[i].size);

	return ( SUCCESS );
}

static int init_flash(void)
{
	return ( flash_init() ? SUCCESS : FAILURE );
}

static int display_flash_config(void)
{
	extern flash_info_t flash_info[];
	int i;
	
	puts("Flash Configuration:\n");
	
	for ( i = 0; i < CFG_MAX_FLASH_BANKS; i++ )
		if ( FLASH_UNKNOWN != flash_info[i].flash_id )
			display_config(i, flash_info[i].phys_base, flash_info[i].size);
	
	return ( SUCCESS );
}

/*
 * Breathe some life into the board...
 *
 * Initialize a serial port as console, and carry out some hardware
 * tests.
 *
 * The first part of initialization is running from Flash memory;
 * its main purpose is to initialize the RAM so that we
 * can relocate the monitor code to RAM.
 */

/*
 * All attempts to come up with a "common" initialization sequence
 * that works for all boards and architectures failed: some of the
 * requirements are just _too_ different. To get rid of the resulting
 * mess of board dependent #ifdef'ed code we now make the whole
 * initialization sequence configurable to the user.
 *
 * The requirements for any new initalization function is simple: it
 * receives a pointer to the "global data" structure as its only
 * argument, and returns an integer return code, where <== 0> means
 * "continue" and <!= 0 means> "fatal error, hang the system".
 */
typedef int (init_fnc_t) (void);

init_fnc_t *init_sequence[] = {
        cpu_init,               /* basic cpu dependent setup             */
        board_init,             /* basic board dependent setup           */
        interrupt_init,         /* set up exceptions                     */
        env_init,               /* initialize environment                */
        init_baudrate,          /* initialze baudrate settings           */
        serial_init,            /* serial communications setup           */
        console_init_f,         /* stage 1 init of console               */
        display_banner,         /* say that we are here                  */
        dram_init,              /* configure available RAM banks         */
        display_dram_config,    /* displays bank number, start, and size */
#if defined(CONFIG_VCMA9)
        checkboard,             /* checks board's identity               */
#endif
        init_flash,             /* initialize flash banks                */
        display_flash_config,   /* display flash information             */
        NULL,                   /* end of table                          */
};

DECLARE_GLOBAL_DATA_PTR;

void start_armboot (void)
{
	init_fnc_t **init_fnc_ptr;
	char *s;
#if defined(CONFIG_VFD) || defined(CONFIG_LCD)
	unsigned long size;
	unsigned long addr;
#endif

	/* Pointer is writable since we allocated a register for it */
	gd = (gd_t*)(_armboot_start - CFG_MALLOC_LEN - sizeof(gd_t));
	/* compiler optimization barrier needed for GCC >= 3.4 */
	__asm__ __volatile__("": : :"memory");

	memset ((void*)gd, 0, sizeof (gd_t));
	gd->bd = (bd_t*)((char*)gd - sizeof(bd_t));
	memset (gd->bd, 0, sizeof (bd_t));

	monitor_flash_len = _bss_start - _armboot_start;

	for (init_fnc_ptr = init_sequence; *init_fnc_ptr; ++init_fnc_ptr) {
		if ((*init_fnc_ptr)() != 0) {
			hang ();
		}
	}

    // Done with init_sequence: 5.
    //
    update_progress_bar();

#ifdef CONFIG_VFD
#	ifndef PAGE_SIZE
#	  define PAGE_SIZE 4096
#	endif
	/*
	 * reserve memory for VFD display (always full pages)
	 */
	/* bss_end is defined in the board-specific linker script */
	addr = (_bss_end + (PAGE_SIZE - 1)) & ~(PAGE_SIZE - 1);
	size = vfd_setmem (addr);
	gd->fb_base = addr;
#endif /* CONFIG_VFD */

#ifdef CONFIG_LCD
#	ifndef PAGE_SIZE
#	  define PAGE_SIZE 4096
#	endif
	/*
	 * reserve memory for LCD display (always full pages)
	 */
	/* bss_end is defined in the board-specific linker script */
	addr = (_bss_end + (PAGE_SIZE - 1)) & ~(PAGE_SIZE - 1);
	size = lcd_setmem (addr);
	gd->fb_base = addr;
#endif /* CONFIG_LCD */

	/* armboot_start is defined in the board-specific linker script */
	mem_malloc_init (_armboot_start - CFG_MALLOC_LEN);

#if (CONFIG_COMMANDS & CFG_CMD_NAND)
	/*
	 * At the moment, this definitely needs to run *after* flash_init()
	 * because flash_init() sets up some things that nand_init() currently
	 * relies upon.
	 */
	puts("NAND Configuration:\n");
	nand_init();
#endif

#if (CONFIG_COMMANDS & CFG_CMD_BML)
	/* 
	 * After we've run flash_init() and nand_init(), run Samsung's
	 * "bad block management" initialization code.
	 */
	BBM_INIT_DEVICE();
#endif

    // Done with Flash: 6.
    //
    update_progress_bar();

#ifdef CONFIG_EINK_APOLLO
	/*
	 * Fiona has an Apollo Eink controller.  We initialize it here.
	 */
	 eink_fb_probe();
#endif

    // Done with Screen: 7.
    //
    update_progress_bar();

#ifdef CONFIG_HAS_DATAFLASH
	AT91F_DataflashInit();
	dataflash_print_info();
#endif

	/* initialize environment */
	env_relocate ();

#ifdef CONFIG_VFD
	/* must do this after the framebuffer is allocated */
	drv_vfd_init();
#endif /* CONFIG_VFD */

	/* IP Address */
	gd->bd->bi_ip_addr = getenv_IPaddr ("ipaddr");

	/* MAC Address */
	{
		int i;
		ulong reg;
		char *s, *e;
		uchar tmp[64];

		i = getenv_r ("ethaddr", tmp, sizeof (tmp));
		s = (i > 0) ? tmp : NULL;

		for (reg = 0; reg < 6; ++reg) {
			gd->bd->bi_enetaddr[reg] = s ? simple_strtoul (s, &e, 16) : 0;
			if (s)
				s = (*e) ? e + 1 : e;
		}
	}

	devices_init ();	/* get the devices list going. */

	jumptable_init ();

	console_init_r ();	/* fully init console as a device */

#if defined(CONFIG_MISC_INIT_R)
	/* miscellaneous platform dependent initialisations */
	misc_init_r ();
#endif

    // Note mark just before enabling interrupts: 8.
    //
    update_progress_bar();

	/* enable exceptions */
	enable_interrupts ();

	/* Perform network card initialisation if necessary */
#ifdef CONFIG_DRIVER_CS8900
	cs8900_get_enetaddr (gd->bd->bi_enetaddr);
#endif

#if defined(CONFIG_DRIVER_SMC91111) || defined (CONFIG_DRIVER_LAN91C96)
	if (getenv ("ethaddr")) {
		smc_set_mac_addr(gd->bd->bi_enetaddr);
	}
#endif /* CONFIG_DRIVER_SMC91111 || CONFIG_DRIVER_LAN91C96 */

	/* Initialize from environment */
	if ((s = getenv ("loadaddr")) != NULL) {
		load_addr = simple_strtoul (s, NULL, 16);
	}
#if (CONFIG_COMMANDS & CFG_CMD_NET)
	if ((s = getenv ("bootfile")) != NULL) {
		copy_filename (BootFile, s, sizeof (BootFile));
	}
#endif	/* CFG_CMD_NET */

#ifdef BOARD_LATE_INIT
	board_late_init ();
#endif

#if defined (CFG_ENV_IS_IN_BBML)
	/*
	 * The BBM code only saves the environment variables to flash if they've
	 * changed.  And, in this case, they should have only "changed" in the
	 * first-time boot case after a full re-flash of the device.
	 */
	 env_save();
#endif

    // Done with board init: 9.
    //
    update_progress_bar();

	/* main_loop() can return to retry autoboot, if so just run it again. */
	for (;;) {
		main_loop ();
	}

	/* NOTREACHED - no way out of command loop except booting */
}

void hang (void)
{
	puts ("### ERROR ### Please RESET the board ###\n");
	for (;;);
}

#ifdef CONFIG_MODEM_SUPPORT
/* called from main loop (common/main.c) */
extern void  dbg(const char *fmt, ...);
int mdm_init (void)
{
	char env_str[16];
	char *init_str;
	int i;
	extern char console_buffer[];
	static inline void mdm_readline(char *buf, int bufsiz);
	extern void enable_putc(void);
	extern int hwflow_onoff(int);

	enable_putc(); /* enable serial_putc() */

#ifdef CONFIG_HWFLOW
	init_str = getenv("mdm_flow_control");
	if (init_str && (strcmp(init_str, "rts/cts") == 0))
		hwflow_onoff (1);
	else
		hwflow_onoff(-1);
#endif

	for (i = 1;;i++) {
		sprintf(env_str, "mdm_init%d", i);
		if ((init_str = getenv(env_str)) != NULL) {
			serial_puts(init_str);
			serial_puts("\n");
			for(;;) {
				mdm_readline(console_buffer, CFG_CBSIZE);
				dbg("ini%d: [%s]", i, console_buffer);

				if ((strcmp(console_buffer, "OK") == 0) ||
					(strcmp(console_buffer, "ERROR") == 0)) {
					dbg("ini%d: cmd done", i);
					break;
				} else /* in case we are originating call ... */
					if (strncmp(console_buffer, "CONNECT", 7) == 0) {
						dbg("ini%d: connect", i);
						return 0;
					}
			}
		} else
			break; /* no init string - stop modem init */

		udelay(100000);
	}

	udelay(100000);

	/* final stage - wait for connect */
	for(;i > 1;) { /* if 'i' > 1 - wait for connection
				  message from modem */
		mdm_readline(console_buffer, CFG_CBSIZE);
		dbg("ini_f: [%s]", console_buffer);
		if (strncmp(console_buffer, "CONNECT", 7) == 0) {
			dbg("ini_f: connected");
			return 0;
		}
	}

	return 0;
}

/* 'inline' - We have to do it fast */
static inline void mdm_readline(char *buf, int bufsiz)
{
	char c;
	char *p;
	int n;

	n = 0;
	p = buf;
	for(;;) {
		c = serial_getc();

		/*		dbg("(%c)", c); */

		switch(c) {
		case '\r':
			break;
		case '\n':
			*p = '\0';
			return;

		default:
			if(n++ > bufsiz) {
				*p = '\0';
				return; /* sanity check */
			}
			*p = c;
			p++;
			break;
		}
	}
}
#endif	/* CONFIG_MODEM_SUPPORT */

#ifdef CONFIG_FIONA
U_BOOT_CMD
(
	kindle, 1, 0, do_fiona_info,
	"kindle  - print info about Kindle's revision\n",
	"- prints revision level and CPU speed\n"
);
#endif

/* vim: set ts=4 sw=4 tw=80: */

