/*
 * (C) Copyright 2000-2002
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>

int display_options (void)
{
	extern char version_string[];

#if defined(BUILD_TAG)
	printf ("\n\n%s, Build: %s\n\n", version_string, BUILD_TAG);
#else
	printf ("\n\n%s\n\n", version_string);
#endif
	return 0;
}

static char print_size_buffer[STRING_MAX];

/*
 * Print sizes as "xxx KB", "xxx.y KB", "xxx MB" or "xxx.y MB" as needed;
 * allow for optional trailing string (like "\n").
 *
 * Changed kB to KB per http://www.netlingo.com/right.cfm?term=kilobyte:
 *
 * A unit of measurement equivalent to one thousand bytes of computer
 * memory or disk capacity. For example, a device that has 256K of memory
 * can store approximately 256,000 bytes (or characters) at one time. In
 * decimal systems, kilo stands for 1,000, but the computer world is
 * based on a binary system of twos instead of tens. So, a kilobyte is
 * actually 1,024 (210) bytes. To distinguish between a decimal K (1,000)
 * and a binary K (1,024), the IEEE has suggested using a small k for a
 * decimal kilo and a capital K for a binary kilo. In international
 * English, outside the U.S.A., the equivalent unit is sometimes seen as
 * "Kb" or "kbyte." 
 */

char *get_size (ulong size, const char *s)
{
	ulong m, n, len;
	ulong d = 1 << 20;		/* 1 MB */
	char  c = 'M';

	if (size < d) {			/* print in KB */
		c = 'K';
		d = 1 << 10;
	}

	n = size / d;

	m = (10 * (size - (n * d)) + (d / 2) ) / d;

	if (m >= 10) {
		m -= 10;
		n += 1;
	}

	sprintf (print_size_buffer,"%ld", n);
	if (m) {
		len = strlen(print_size_buffer);
		sprintf (&print_size_buffer[len], ".%ld", m);
	}
	len = strlen(print_size_buffer);
	sprintf (&print_size_buffer[len], " %cB%s", c, s);
	
	return print_size_buffer;
}

void print_size (ulong size, const char *s)
{
	get_size (size, s);
	puts (print_size_buffer);
}

